#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""
This script dynamically builds a workflow tailored to the size of the communicator.
"""

# This is the maximum random time in seconds for every process to sleep before starting.
# This should be long enough to avoid filesystem contention when loading shared
# libraries and python modules, but short enough that the scheduler does not think that
# MPI has hung and kills the job.
startup_delay = 2.0

# These are builtin modules, hopefully fast to load.
import random
import time

wait = random.uniform(0.0, startup_delay)
time.sleep(wait)

from toast.mpi import MPI

# Now import the remaining modules
import os
import sys
import re
import copy
import argparse
import traceback
import pickle

from datetime import datetime

import psutil

import numpy as np

import healpy as hp

from toast.mpi import get_world, Comm

from toast.dist import distribute_uniform, Data, distribute_discrete

from toast.utils import Logger, Environment, memreport

from toast.timing import function_timer, GlobalTimers, Timer, gather_timers
from toast.timing import dump as dump_timing

from toast.todmap import TODGround

from toast import pipeline_tools

from toast import rng

from toast.pipeline_tools import Focalplane

from toast.tod.sim_focalplane import hex_pol_angles_qu, hex_layout, hex_nring

from toast.schedule import run_scheduler


# This function is part of the package in TOAST 3.0.  Remove this here when porting.


def fake_hexagon_focalplane(
    n_pix=7,
    width_deg=5.0,
    samplerate=1.0,
    epsilon=0.0,
    net=1.0,
    fmin=0.0,
    alpha=1.0,
    fknee=0.05,
):
    pol_A = hex_pol_angles_qu(n_pix, offset=0.0)
    pol_B = hex_pol_angles_qu(n_pix, offset=90.0)
    quat_A = hex_layout(n_pix, width_deg, "D", "A", pol_A)
    quat_B = hex_layout(n_pix, width_deg, "D", "B", pol_B)

    det_data = dict(quat_A)
    det_data.update(quat_B)

    nrings = hex_nring(n_pix)
    detfwhm = 0.5 * 60.0 * width_deg / (2 * nrings - 1)

    for det in det_data.keys():
        det_data[det]["pol_leakage"] = epsilon
        det_data[det]["fmin"] = fmin
        det_data[det]["fknee"] = fknee
        det_data[det]["alpha"] = alpha
        det_data[det]["NET"] = net
        det_data[det]["fwhm_arcmin"] = detfwhm
        det_data[det]["fsample"] = samplerate

    return Focalplane(detector_data=det_data, sample_rate=samplerate)


def get_node_mem(mpicomm, node_rank):
    avail = 2 ** 62
    if node_rank == 0:
        vmem = psutil.virtual_memory()._asdict()
        avail = vmem["available"]
    if mpicomm is not None:
        avail = mpicomm.allreduce(avail, op=MPI.MIN)
    return int(avail)


def sample_distribution(
    rank, procs_per_node, bytes_per_node, total_samples, sample_rate
):
    # For this benchmark, we start by ramping up to a realistic number of detectors for
    # one day of data. Then we extend the timespan to achieve the desired number of
    # samples.

    log = Logger.get()

    # Hex-packed 127 pixels (6 rings) times two dets per pixel.
    # max_detector = 254

    # Hex-packed 1027 pixels (18 rings) times two dets per pixel.
    max_detector = 2054

    # Minimum time span (one day)
    min_time_samples = int(24 * 3600 * sample_rate)

    # For the minimum time span, scale up the number of detectors to reach the
    # requested total sample size.

    n_detector = 1
    test_samples = n_detector * min_time_samples

    while test_samples < total_samples and n_detector < max_detector:
        n_detector += 1
        test_samples = n_detector * min_time_samples

    if rank == 0:
        log.debug(
            "  Dist total = {}, using {} detectors at min time samples = {}".format(
                total_samples, n_detector, min_time_samples
            )
        )

    # For this number of detectors, determine the group size needed to fit the
    # minimum number of samples in memory.  In practice, one day will actually be
    # split up into multiple observations.  However, sizing the groups this way ensures
    # that each group will have multiple observations and improve the load balancing.

    det_bytes_per_sample = 2 * (  # At most 2 detector data copies.
        8  # 64 bit float / ints used
        * (1 + 4)  # detector timestream  # pixel index and 3 IQU weights
        + 1  # one byte per sample for flags
    )

    common_bytes_per_sample = (
        8 * (4)  # 64 bit floats  # One quaternion per sample
        + 1  # one byte per sample for common flag
    )

    group_nodes = 0
    group_mem = 0.0

    # This just ensures we go through the loop once.
    min_time_mem = 1.0

    while group_mem < min_time_mem:
        group_nodes += 1
        group_procs = group_nodes * procs_per_node
        group_mem = group_nodes * bytes_per_node

        # NOTE:  change this when moving to toast-3, since common data is in shared mem.
        # So the prefactor should be nodes per group, not group_procs.
        bytes_per_samp = (
            n_detector * det_bytes_per_sample + group_procs * common_bytes_per_sample
        )
        # bytes_per_samp = (
        #     n_detector * det_bytes_per_sample + group_nodes * common_bytes_per_sample
        # )
        min_time_mem = min_time_samples * bytes_per_samp
        if rank == 0:
            log.verbose(
                "  Dist testing {} group nodes, {} proc/node, group mem = {}, comparing to minimum = {} ({} samp * {} bytes/samp)".format(
                    group_nodes,
                    procs_per_node,
                    group_mem,
                    min_time_mem,
                    min_time_samples,
                    bytes_per_samp,
                )
            )

    if rank == 0:
        log.debug("  Dist selecting {} nodes per group".format(group_nodes))

    # Now set the number of groups to get the target number of total samples.

    group_time_samples = min_time_samples
    group_samples = n_detector * group_time_samples

    n_group = 1 + (total_samples // group_samples)

    time_samples = n_group * group_time_samples

    if rank == 0:
        log.debug(
            "  Dist using {} groups, each with {} / {} (time / total) samples".format(
                n_group, group_time_samples, group_samples
            )
        )
        log.debug("  Dist using {} total samples".format(n_detector * time_samples))

    return (
        n_detector,
        time_samples,
        group_procs,
        group_nodes,
        n_group,
        group_time_samples,
    )


def job_size(mpicomm):
    log = Logger.get()

    procs_per_node = 1
    node_rank = 0
    nodecomm = None
    rank = 0
    procs = 1

    if mpicomm is not None:
        rank = mpicomm.rank
        procs = mpicomm.size
        nodecomm = mpicomm.Split_type(MPI.COMM_TYPE_SHARED, 0)
        node_rank = nodecomm.rank
        procs_per_node = nodecomm.size
        min_per_node = mpicomm.allreduce(procs_per_node, op=MPI.MIN)
        max_per_node = mpicomm.allreduce(procs_per_node, op=MPI.MAX)
        if min_per_node != max_per_node:
            raise RuntimeError("Nodes have inconsistent numbers of MPI ranks")

    # One process on each node gets available RAM and communicates it
    avail = get_node_mem(mpicomm, node_rank)

    n_node = procs // procs_per_node

    if rank == 0:
        log.info(
            "Job running on {} nodes each with {} processes ({} total)".format(
                n_node, procs_per_node, procs
            )
        )
    return (procs_per_node, avail)


def job_config(mpicomm, cases):
    env = Environment.get()
    log = Logger.get()

    class args:
        debug = False
        # TOD Ground options
        el_mod_step_deg = 0.0
        el_mod_rate_hz = 0.0
        el_mod_amplitude_deg = 1.0
        el_mod_sine = False
        el_nod_deg = False
        el_nod_every_scan = False
        start_with_el_nod = False
        end_with_el_nod = False
        scan_rate = 1.0
        scan_rate_el = 0.0
        scan_accel = 1.0
        scan_accel_el = 0.0
        scan_cosecant_modulate = False
        sun_angle_min = 30.0
        schedule = None  # required
        weather = "SIM"
        timezone = 0
        sample_rate = 100.0
        coord = "C"
        split_schedule = None
        sort_schedule = False
        hwp_rpm = 10.0
        hwp_step_deg = None
        hwp_step_time_s = None
        elevation_noise_a = 0.0
        elevation_noise_b = 0.0
        freq = "150"
        do_daymaps = False
        do_seasonmaps = False
        # Pointing options
        nside = 1024
        nside_submap = 16
        single_precision_pointing = False
        common_flag_mask = 1
        # Polyfilter options
        apply_polyfilter = False
        poly_order = 0
        # Ground filter options
        apply_groundfilter = False
        ground_order = 0
        # Atmosphere options
        simulate_atmosphere = False
        simulate_coarse_atmosphere = False
        focalplane_radius_deg = None
        atm_verbosity = 0
        atm_lmin_center = 0.01
        atm_lmin_sigma = 0.001
        atm_lmax_center = 10.0
        atm_lmax_sigma = 10.0
        atm_gain = 2.0e-5
        atm_gain_coarse = 8.0e-5
        atm_zatm = 40000.0
        atm_zmax = 200.0
        atm_xstep = 10.0
        atm_ystep = 10.0
        atm_zstep = 10.0
        atm_nelem_sim_max = 10000
        atm_wind_dist = 3000.0
        atm_z0_center = 2000.0
        atm_z0_sigma = 0.0
        atm_T0_center = 280.0
        atm_T0_sigma = 10.0
        atm_cache = None
        atm_apply_flags = False
        # Noise simulation options
        simulate_noise = False
        # Gain scrambler
        apply_gainscrambler = False
        gain_sigma = 0.01
        # Map maker
        mapmaker_prefix = "toast"
        mapmaker_mask = None
        mapmaker_weightmap = None
        mapmaker_iter_max = 20
        mapmaker_precond_width = 100
        mapmaker_prefilter_order = None
        mapmaker_baseline_length = 200.0
        mapmaker_noisefilter = False
        mapmaker_fourier2D_order = None
        mapmaker_fourier2D_subharmonics = None
        write_hits = True
        write_binmap = True
        write_wcov = False
        write_wcov_inv = False
        zip_maps = False
        # Monte Carlo
        MC_start = 0
        MC_count = 1
        # Sky signal
        input_map = None
        simulate_sky = True
        # Input dir
        auxdir = "toast_inputs"
        # Output
        outdir = "toast"
        tidas = None
        spt3g = None

    parser = argparse.ArgumentParser(
        description="Run a TOAST workflow scaled appropriately to the MPI communicator size and available memory.",
        fromfile_prefix_chars="@",
    )

    parser.add_argument(
        "--node_mem_gb",
        required=False,
        default=None,
        type=float,
        help="Use this much memory per node in GB",
    )

    parser.add_argument(
        "--dry_run",
        required=False,
        default=None,
        type=str,
        help="Comma-separated total_procs,node_procs to simulate.",
    )

    parser.parse_args(namespace=args)

    procs = 1
    rank = 0
    if mpicomm is not None:
        procs = mpicomm.size
        rank = mpicomm.rank

    avail_node_bytes = None
    procs_per_node = None

    if args.dry_run is not None:
        dryrun_total, dryrun_node = args.dry_run.split(",")
        dryrun_total = int(dryrun_total)
        dryrun_node = int(dryrun_node)
        if rank == 0:
            log.info(
                "DRY RUN simulating {} total processes with {} per node".format(
                    dryrun_total, dryrun_node
                )
            )
        procs_per_node = dryrun_node
        procs = dryrun_total
        # We are simulating the distribution
        avail_node_bytes = get_node_mem(mpicomm, 0)

    else:
        # Get information about the actual job size
        procs_per_node, avail_node_bytes = job_size(mpicomm)

    if rank == 0:
        log.info(
            "Minimum detected per-node memory available is {:0.2f} GB".format(
                avail_node_bytes / (1024 ** 3)
            )
        )

    if args.node_mem_gb is not None:
        avail_node_bytes = int((1024 ** 3) * args.node_mem_gb)
        if rank == 0:
            log.info(
                "Setting per-node available memory to {:0.2f} GB as requested".format(
                    avail_node_bytes / (1024 ** 3)
                )
            )

    # Based on the total number of processes and count per node, choose the number of
    # nodes in each observation and a focalplane such that every process has >= 4
    # detectors.

    n_nodes = procs // procs_per_node
    if rank == 0:
        log.info("Job has {} total nodes".format(n_nodes))

    if rank == 0:
        log.info("Examining {} possible cases to run:".format(len(cases)))

    selected_case = None
    selected_nodes = None
    n_detector = None
    time_samples = None
    group_procs = None
    group_nodes = None
    n_group = None
    group_time_samples = None

    for case_name, case_samples in cases.items():
        (
            case_n_detector,
            case_time_samples,
            case_group_procs,
            case_group_nodes,
            case_n_group,
            case_group_time_samples,
        ) = sample_distribution(
            rank, procs_per_node, avail_node_bytes, case_samples, args.sample_rate
        )

        case_min_nodes = case_n_group * case_group_nodes
        if rank == 0:
            log.info(
                "  {:8s}: requires {:d} nodes for {} MPI ranks and {:0.1f}GB per node".format(
                    case_name,
                    case_min_nodes,
                    procs_per_node,
                    avail_node_bytes / (1024 ** 3),
                )
            )

        if selected_nodes is None:
            if case_min_nodes <= n_nodes:
                # First case that fits in our job
                selected_case = case_name
                selected_nodes = case_min_nodes
                n_detector = case_n_detector
                time_samples = case_time_samples
                group_procs = case_group_procs
                group_nodes = case_group_nodes
                n_group = case_n_group
                group_time_samples = case_group_time_samples
        else:
            if (case_min_nodes <= n_nodes) and (case_min_nodes >= selected_nodes):
                # This case fits in our job and is larger than the current one
                selected_case = case_name
                selected_nodes = case_min_nodes
                n_detector = case_n_detector
                time_samples = case_time_samples
                group_procs = case_group_procs
                group_nodes = case_group_nodes
                n_group = case_n_group
                group_time_samples = case_group_time_samples

    if selected_case is None:
        msg = (
            "None of the available cases fit into aggregate memory.  Use a larger job."
        )
        if rank == 0:
            log.error(msg)
        raise RuntimeError(msg)
    else:
        if rank == 0:
            log.info("Selected case '{}'".format(selected_case))

    if rank == 0:
        log.info("Using groups of {} nodes".format(group_nodes))

    # Adjust number of groups

    if n_nodes % group_nodes != 0:
        msg = "Current number of nodes ({}) is not divisible by the required group size ({})".format(
            n_nodes, group_nodes
        )
        if rank == 0:
            log.error(msg)
        raise RuntimeError(msg)

    n_group = n_nodes // group_nodes
    group_time_samples = 1 + time_samples // n_group

    group_seconds = group_time_samples / args.sample_rate

    if args.simulate_atmosphere and args.weather is None:
        raise RuntimeError("Cannot simulate atmosphere without a TOAST weather file")

    comm = None
    if mpicomm is None or args.dry_run is not None:
        comm = Comm(world=None)
    else:
        comm = Comm(world=mpicomm, groupsize=group_procs)

    jobdate = datetime.now().strftime("%Y%m%d-%H:%M:%S")

    args.outdir += "_{:06d}_grp-{:04d}p-{:02d}n_{}".format(
        procs, group_procs, group_nodes, jobdate
    )
    args.auxdir = os.path.join(args.outdir, "inputs")

    if rank == 0:
        os.makedirs(args.outdir)
        os.makedirs(args.auxdir, exist_ok=True)

    if rank == 0:
        with open(os.path.join(args.outdir, "log"), "w") as f:
            f.write("Running at {}\n".format(jobdate))
            f.write("TOAST version = {}\n".format(env.version()))
            f.write("TOAST max threads = {}\n".format(env.max_threads()))
            f.write("MPI Processes = {}\n".format(procs))
            f.write("MPI Processes per node = {}\n".format(procs_per_node))
            f.write(
                "Memory per node = {:0.2f} GB\n".format(avail_node_bytes / (1024 ** 3))
            )
            f.write("Number of groups = {}\n".format(n_group))
            f.write("Group nodes = {}\n".format(group_nodes))
            f.write("Group MPI Processes = {}\n".format(group_procs))
            f.write("Case selected = {}\n".format(selected_case))
            f.write("Case number of detectors = {}\n".format(n_detector))
            f.write(
                "Case total samples = {}\n".format(
                    n_group * group_time_samples * n_detector
                )
            )
            f.write(
                "Case samples per group = {}\n".format(group_time_samples * n_detector)
            )
            f.write("Case data seconds per group = {}\n".format(group_seconds))
            f.write("Parameters:\n")
            for k, v in vars(args).items():
                if re.match(r"_.*", k) is None:
                    f.write("  {} = {}\n".format(k, v))

    args.schedule = os.path.join(args.auxdir, "schedule.txt")
    args.input_map = os.path.join(args.auxdir, "cmb.fits")

    return args, comm, n_nodes, n_detector, selected_case, group_seconds, n_group


def create_schedules(args, max_ces_seconds, days):
    opts = [
        "--site-lat",
        str(-22.958064),
        "--site-lon",
        str(-67.786222),
        "--site-alt",
        str(5200.0),
        "--site-name",
        "ATACAMA",
        "--telescope",
        "atacama_telescope",
        "--patch-coord",
        "C",
        "--el-min-deg",
        str(30.0),
        "--el-max-deg",
        str(80.0),
        "--sun-el-max-deg",
        str(90.0),
        "--sun-avoidance-angle-deg",
        str(30.0),
        "--moon-avoidance-angle-deg",
        str(10.0),
        "--start",
        "2021-06-01 00:00:00",
        "--gap-s",
        str(600.0),
        "--gap-small-s",
        str(0.0),
        "--fp-radius-deg",
        str(0.0),
        "--patch",
        "BICEP,1,-10,-55,10,-58",
        "--ces-max-time-s",
        str(max_ces_seconds),
        "--operational-days",
        str(days),
    ]

    if not os.path.isfile(args.schedule):
        log = Logger.get()
        log.info("Generating input schedule file {}:".format(args.schedule))

        opts.extend(["--out", args.schedule])
        run_scheduler(opts=opts)


def create_input_maps(args):
    if not os.path.isfile(args.input_map):
        log = Logger.get()
        log.info("Generating input map {}".format(args.input_map))

        # This is *completely* fake- just to have something on the sky besides zeros.

        ell = np.arange(3 * args.nside - 1, dtype=np.float64)

        sig = 50.0
        numer = ell - 30.0
        tspec = (1.0 / (sig * np.sqrt(2.0 * np.pi))) * np.exp(
            -0.5 * numer ** 2 / sig ** 2
        )
        tspec *= 2000.0

        sig = 100.0
        numer = ell - 500.0
        espec = (1.0 / (sig * np.sqrt(2.0 * np.pi))) * np.exp(
            -0.5 * numer ** 2 / sig ** 2
        )
        espec *= 1.0

        cls = (
            tspec,
            espec,
            np.zeros(3 * args.nside - 1, dtype=np.float32),
            np.zeros(3 * args.nside - 1, dtype=np.float32),
        )
        maps = hp.synfast(
            cls,
            args.nside,
            pol=True,
            pixwin=False,
            sigma=None,
            new=True,
            fwhm=np.radians(3.0 / 60.0),
            verbose=False,
        )
        # for m in maps:
        #     hp.reorder(m, inp="RING", out="NEST")
        hp.write_map(args.input_map, maps, nest=True, fits_IDL=False, dtype=np.float32)

        import matplotlib.pyplot as plt

        hp.mollview(maps[0])
        plt.savefig("{}_fake-T.png".format(args.input_map))
        plt.close()

        hp.mollview(maps[1])
        plt.savefig("{}_fake-E.png".format(args.input_map))
        plt.close()


@function_timer
def create_focalplanes(args, comm, schedules, n_detector):
    """Attach a focalplane to each of the schedules.

    Args:
        schedules (list) :  List of Schedule instances.
            Each schedule has two members, telescope
            and ceslist, a list of CES objects.
    Returns:
        detweights (dict) : Inverse variance noise weights for every
            detector across all focal planes. In [K_CMB^-2].
            They can be used to bin the TOD.
    """

    # Load focalplane information

    focalplanes = []
    if comm.world_rank == 0:
        # First create a fake hexagonal focalplane
        n_pixel = 1
        ring = 1
        while 2 * n_pixel < n_detector:
            n_pixel += 6 * ring
            ring += 1

        fake = fake_hexagon_focalplane(
            n_pix=n_pixel,
            width_deg=10.0,
            samplerate=100.0,
            epsilon=0.0,
            net=10.0,
            fmin=0.0,
            alpha=1.0,
            fknee=0.05,
        )

        # Now truncate the detectors to the desired count.
        newdat = dict()
        off = 0
        for k, v in fake.detector_data.items():
            newdat[k] = v
            off += 1
            if off >= n_detector:
                break
        fake.detector_data = newdat
        fake.reset_properties()
        focalplanes.append(fake)

    if comm.comm_world is not None:
        focalplanes = comm.comm_world.bcast(focalplanes)

    if len(focalplanes) == 1 and len(schedules) > 1:
        focalplanes *= len(schedules)

    # Append a focal plane and telescope to each entry in the schedules
    # list and assemble a detector weight dictionary that represents all
    # detectors in all focalplanes
    detweights = {}
    for schedule, focalplane in zip(schedules, focalplanes):
        schedule.telescope.focalplane = focalplane
        detweights.update(schedule.telescope.focalplane.detweights)

    return detweights


@function_timer
def create_observation(args, comm, telescope, ces, verbose=True):
    """Create a TOAST observation.

    Create an observation for the CES scan

    Args:
        args :  argparse arguments
        comm :  TOAST communicator
        ces (CES) :  One constant elevation scan

    """
    focalplane = telescope.focalplane
    site = telescope.site
    weather = site.weather
    noise = focalplane.noise
    totsamples = int((ces.stop_time - ces.start_time) * args.sample_rate)

    # create the TOD for this observation

    if comm.comm_group is not None:
        ndetrank = comm.comm_group.size
    else:
        ndetrank = 1

    if args.el_nod_deg and (ces.subscan == 0 or args.el_nod_every_scan):
        el_nod = args.el_nod_deg
    else:
        el_nod = None

    try:
        tod = TODGround(
            comm.comm_group,
            focalplane.detquats,
            totsamples,
            detranks=ndetrank,
            boresight_angle=ces.boresight_angle,
            firsttime=ces.start_time,
            rate=args.sample_rate,
            site_lon=site.lon,
            site_lat=site.lat,
            site_alt=site.alt,
            azmin=ces.azmin,
            azmax=ces.azmax,
            el=ces.el,
            el_nod=el_nod,
            start_with_elnod=args.start_with_el_nod,
            end_with_elnod=args.end_with_el_nod,
            el_mod_step=args.el_mod_step_deg,
            el_mod_rate=args.el_mod_rate_hz,
            el_mod_amplitude=args.el_mod_amplitude_deg,
            el_mod_sine=args.el_mod_sine,
            scanrate=args.scan_rate,
            scanrate_el=args.scan_rate_el,
            scan_accel=args.scan_accel,
            scan_accel_el=args.scan_accel_el,
            cosecant_modulation=args.scan_cosecant_modulate,
            CES_start=None,
            CES_stop=None,
            sun_angle_min=args.sun_angle_min,
            coord=args.coord,
            sampsizes=None,
            report_timing=args.debug,
            hwprpm=args.hwp_rpm,
            hwpstep=args.hwp_step_deg,
            hwpsteptime=args.hwp_step_time_s,
        )
    except RuntimeError as e:
        raise RuntimeError(
            'Failed to create TOD for {}-{}-{}: "{}"'
            "".format(ces.name, ces.scan, ces.subscan, e)
        )

    # Create the observation

    obs = {}
    obs["name"] = "CES-{}-{}-{}-{}-{}".format(
        site.name, telescope.name, ces.name, ces.scan, ces.subscan
    )
    obs["tod"] = tod
    obs["baselines"] = None
    obs["noise"] = copy.deepcopy(noise)
    obs["id"] = int(ces.mjdstart * 10000)
    obs["intervals"] = tod.subscans
    obs["site"] = site
    obs["site_name"] = site.name
    obs["site_id"] = site.id
    obs["altitude"] = site.alt
    obs["weather"] = site.weather
    obs["telescope"] = telescope
    obs["telescope_name"] = telescope.name
    obs["telescope_id"] = telescope.id
    obs["focalplane"] = telescope.focalplane.detector_data
    obs["fpradius"] = telescope.focalplane.radius
    obs["start_time"] = ces.start_time
    obs["season"] = ces.season
    obs["date"] = ces.start_date
    obs["MJD"] = ces.mjdstart
    obs["rising"] = ces.rising
    obs["mindist_sun"] = ces.mindist_sun
    obs["mindist_moon"] = ces.mindist_moon
    obs["el_sun"] = ces.el_sun
    return obs


@function_timer
def create_observations(args, comm, schedules):
    """Create and distribute TOAST observations for every CES in
    schedules.

    Args:
        schedules (iterable) :  a list of Schedule objects.
    """
    log = Logger.get()
    timer = Timer()
    timer.start()

    data = Data(comm)

    # Loop over the schedules, distributing each schedule evenly across
    # the process groups.  For now, we'll assume that each schedule has
    # the same number of operational days and the number of process groups
    # matches the number of operational days.  Relaxing these constraints
    # will cause the season break to occur on different process groups
    # for different schedules and prevent splitting the communicator.

    total_samples = 0
    group_samples = 0
    for schedule in schedules:

        telescope = schedule.telescope
        all_ces = schedule.ceslist
        nces = len(all_ces)

        breaks = pipeline_tools.get_breaks(comm, all_ces, args)

        ces_weights = [x.stop_time - x.start_time for x in all_ces]
        groupdist = distribute_discrete(ces_weights, comm.ngroups, breaks=breaks)

        # groupdist = distribute_uniform(nces, comm.ngroups, breaks=breaks)
        group_firstobs = groupdist[comm.group][0]
        group_numobs = groupdist[comm.group][1]

        for ices in range(group_firstobs, group_firstobs + group_numobs):
            obs = create_observation(args, comm, telescope, all_ces[ices])
            group_samples += obs["tod"].total_samples
            data.obs.append(obs)

    if comm.comm_rank is not None:
        if comm.comm_group.rank == 0:
            total_samples = comm.comm_rank.allreduce(group_samples, op=MPI.SUM)
        total_samples = comm.comm_group.bcast(total_samples, root=0)
    if comm.comm_world is None or comm.comm_group.rank == 0:
        log.info("Group # {:4} has {} observations.".format(comm.group, len(data.obs)))

    if len(data.obs) == 0:
        raise RuntimeError(
            "Too many tasks. Every MPI task must "
            "be assigned to at least one observation."
        )

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0:
        timer.report("Simulated scans")

    # Split the data object for each telescope for separate mapmaking.
    # We could also split by site.

    if len(schedules) > 1:
        telescope_data = data.split("telescope_name")
        if len(telescope_data) == 1:
            # Only one telescope available
            telescope_data = []
    else:
        telescope_data = []
    telescope_data.insert(0, ("all", data))
    return data, telescope_data, total_samples


def setup_sigcopy(args):
    """Determine if an extra copy of the atmospheric signal is needed.

    When we simulate multichroic focal planes, the frequency-independent
    part of the atmospheric noise is simulated first and then the
    frequency scaling is applied to a copy of the atmospheric noise.
    """
    if len(args.freq.split(",")) == 1:
        totalname = "total"
        totalname_freq = "total"
    else:
        totalname = "total"
        totalname_freq = "total_freq"

    return totalname, totalname_freq


@function_timer
def setup_output(args, comm, mc, freq):
    outpath = "{}/{:03}/{:03}".format(args.outdir, mc, int(freq))
    if comm.world_rank == 0:
        os.makedirs(outpath, exist_ok=True)
    return outpath


def main():
    env = Environment.get()
    env.enable_function_timers()

    log = Logger.get()
    gt = GlobalTimers.get()
    gt.start("toast_benchmark (total)")

    mpiworld, procs, rank = get_world()

    if rank == 0:
        log.info("TOAST version = {}".format(env.version()))
        log.info("Using a maximum of {} threads per process".format(env.max_threads()))
    if mpiworld is None:
        log.info("Running serially with one process at {}".format(str(datetime.now())))
    else:
        if rank == 0:
            log.info(
                "Running with {} processes at {}".format(procs, str(datetime.now()))
            )

    cases = {
        "tiny": 5000000,  # O(1) GB RAM
        "xsmall": 50000000,  # O(10) GB RAM
        "small": 500000000,  # O(100) GB RAM
        "medium": 5000000000,  # O(1) TB RAM
        "large": 50000000000,  # O(10) TB RAM
        "xlarge": 500000000000,  # O(100) TB RAM
        "heroic": 5000000000000,  # O(1000) TB RAM
    }

    args, comm, n_nodes, n_detector, case, group_seconds, n_group = job_config(
        mpiworld, cases
    )

    # Note:  The number of "days" here will just be an approximation of the desired
    # data volume since we are doing a realistic schedule for a real observing site.

    n_days = int(2.0 * (group_seconds * n_group) / (24 * 3600))
    if n_days == 0:
        n_days = 1

    if rank == 0:
        log.info(
            "Using {} detectors for approximately {} days".format(n_detector, n_days)
        )

    # Create the schedule file and input maps on one process
    if rank == 0:
        create_schedules(args, group_seconds, n_days)
        create_input_maps(args)
    if mpiworld is not None:
        mpiworld.barrier()

    if args.dry_run is not None:
        if rank == 0:
            log.info("Exit from dry run")
        # We are done!
        sys.exit(0)

    gt.start("toast_benchmark (science work)")

    # Load and broadcast the schedule file

    schedules = pipeline_tools.load_schedule(args, comm)

    # Load the weather and append to schedules

    pipeline_tools.load_weather(args, comm, schedules)

    # Simulate the focalplane

    detweights = create_focalplanes(args, comm, schedules, n_detector)

    # Create the TOAST data object to match the schedule.  This will
    # include simulating the boresight pointing.

    data, telescope_data, total_samples = create_observations(args, comm, schedules)

    # handle = None
    # if comm.world_rank == 0:
    #     handle = open(os.path.join(args.outdir, "distdata.txt"), "w")
    # data.info(handle)
    # if comm.world_rank == 0:
    #     handle.close()
    # if comm.comm_world is not None:
    #     comm.comm_world.barrier()

    # Split the communicator for day and season mapmaking

    time_comms = pipeline_tools.get_time_communicators(args, comm, data)

    # Expand boresight quaternions into detector pointing weights and
    # pixel numbers

    pipeline_tools.expand_pointing(args, comm, data)

    # Optionally rewrite the noise PSD:s in each observation to include
    # elevation-dependence

    pipeline_tools.get_elevation_noise(args, comm, data)

    # Purge the pointing if we are NOT going to export the
    # data to a TIDAS volume
    if (args.tidas is None) and (args.spt3g is None):
        for ob in data.obs:
            tod = ob["tod"]
            tod.free_radec_quats()

    # Prepare auxiliary information for distributed map objects

    signalname = pipeline_tools.scan_sky_signal(args, comm, data, "signal")

    # Set up objects to take copies of the TOD at appropriate times

    totalname, totalname_freq = setup_sigcopy(args)

    # Loop over Monte Carlos

    firstmc = args.MC_start
    nsimu = args.MC_count

    freqs = [float(freq) for freq in args.freq.split(",")]
    nfreq = len(freqs)

    for mc in range(firstmc, firstmc + nsimu):

        pipeline_tools.simulate_atmosphere(args, comm, data, mc, totalname)

        # Loop over frequencies with identical focal planes and identical
        # atmospheric noise.

        for ifreq, freq in enumerate(freqs):

            if comm.world_rank == 0:
                log.info(
                    "Processing frequency {}GHz {} / {}, MC = {}".format(
                        freq, ifreq + 1, nfreq, mc
                    )
                )

            # Make a copy of the atmosphere so we can scramble the gains and apply
            # frequency-dependent scaling.
            pipeline_tools.copy_signal(args, comm, data, totalname, totalname_freq)

            pipeline_tools.scale_atmosphere_by_frequency(
                args, comm, data, freq=freq, mc=mc, cache_name=totalname_freq
            )

            pipeline_tools.update_atmospheric_noise_weights(args, comm, data, freq, mc)

            # Add previously simulated sky signal to the atmospheric noise.

            pipeline_tools.add_signal(
                args, comm, data, totalname_freq, signalname, purge=(nsimu == 1)
            )

            mcoffset = ifreq * 1000000

            pipeline_tools.simulate_noise(
                args, comm, data, mc + mcoffset, totalname_freq
            )

            pipeline_tools.scramble_gains(
                args, comm, data, mc + mcoffset, totalname_freq
            )

            outpath = setup_output(args, comm, mc + mcoffset, freq)

            # Bin and destripe maps

            pipeline_tools.apply_mapmaker(
                args,
                comm,
                data,
                outpath,
                totalname_freq,
                time_comms=time_comms,
                telescope_data=telescope_data,
                first_call=(mc == firstmc),
            )

            if args.apply_polyfilter or args.apply_groundfilter:

                # Filter signal

                pipeline_tools.apply_polyfilter(args, comm, data, totalname_freq)

                pipeline_tools.apply_groundfilter(args, comm, data, totalname_freq)

                # Bin filtered maps

                pipeline_tools.apply_mapmaker(
                    args,
                    comm,
                    data,
                    outpath,
                    totalname_freq,
                    time_comms=time_comms,
                    telescope_data=telescope_data,
                    first_call=False,
                    extra_prefix="filtered",
                    bin_only=True,
                )

    gt.stop_all()
    if mpiworld is not None:
        mpiworld.barrier()

    runtime = gt.seconds("toast_benchmark (science work)")
    prefactor = 1.0e-3
    kilo_samples = 1.0e-3 * total_samples
    sample_factor = 1.2
    det_factor = 2.0
    metric = (
        prefactor
        * n_detector ** det_factor
        * kilo_samples ** sample_factor
        / (n_nodes * runtime)
    )
    if rank == 0:
        msg = "Science Metric: {:0.1e} * ({:d}**{:0.2f}) * ({:0.3e}**{:0.3f}) / ({:0.1f} * {}) = {:0.2f}".format(
            prefactor,
            n_detector,
            det_factor,
            kilo_samples,
            sample_factor,
            runtime,
            n_nodes,
            metric,
        )
        log.info("")
        log.info(msg)
        log.info("")
        with open(os.path.join(args.outdir, "log"), "a") as f:
            f.write(msg)
            f.write("\n\n")

    timer = Timer()
    timer.start()
    alltimers = gather_timers(comm=mpiworld)
    if comm.world_rank == 0:
        out = os.path.join(args.outdir, "timing")
        dump_timing(alltimers, out)
        with open(os.path.join(args.outdir, "log"), "a") as f:
            f.write("Copy of Global Timers:\n")
            with open("{}.csv".format(out), "r") as t:
                f.write(t.read())
        timer.stop()
        timer.report("Gather and dump timing info")
    return


if __name__ == "__main__":
    try:
        main()
    except Exception:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        if procs == 1:
            raise
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None:
            mpiworld.Abort(6)



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..tod import AnalyticNoise, OpSimNoise


def add_noise_args(parser):
    """Add the noise simulation arguments"""
    parser.add_argument(
        "--noise",
        required=False,
        action="store_true",
        help="Add simulated noise",
        dest="simulate_noise",
    )
    parser.add_argument(
        "--simulate-noise",
        required=False,
        action="store_true",
        help="Add simulated noise",
        dest="simulate_noise",
    )
    parser.add_argument(
        "--no-noise",
        required=False,
        action="store_false",
        help="Do not add simulated noise",
        dest="simulate_noise",
    )
    parser.add_argument(
        "--no-simulate-noise",
        required=False,
        action="store_false",
        help="Do not add simulated noise",
        dest="simulate_noise",
    )
    parser.set_defaults(simulate_noise=False)
    # `sample-rate` may be already added
    try:
        parser.add_argument(
            "--sample-rate",
            required=False,
            default=100.0,
            type=np.float,
            help="Detector sample rate (Hz)",
        )
    except argparse.ArgumentError:
        pass
    return


@function_timer
def simulate_noise(
    args, comm, data, mc, cache_prefix=None, verbose=True, overwrite=False
):
    """Simulate electronic noise"""
    if not args.simulate_noise:
        return
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0 and verbose:
        log.info("Simulating noise")
    if overwrite:
        # Clear existing signal from the cache
        for obs in data.obs:
            tod = obs["tod"]
            if cache_prefix is None:
                prefix = tod.SIGNAL_NAME
            else:
                prefix = cache_prefix
            tod.cache.clear(prefix + "_.*")
    nse = OpSimNoise(out=cache_prefix, realization=mc)
    nse.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Simulate noise")
    return


@function_timer
def get_analytic_noise(args, comm, focalplane, verbose=True):
    """Create a TOAST noise object.

    Create a noise object from the 1/f noise parameters contained in the
    focalplane database.

    """
    timer = Timer()
    timer.start()
    detectors = sorted(focalplane.keys())
    fmin = {}
    fknee = {}
    alpha = {}
    NET = {}
    rates = {}
    for d in detectors:
        rates[d] = args.sample_rate
        fmin[d] = focalplane[d]["fmin"]
        fknee[d] = focalplane[d]["fknee"]
        alpha[d] = focalplane[d]["alpha"]
        NET[d] = focalplane[d]["NET"]
    noise = AnalyticNoise(
        rate=rates, fmin=fmin, detectors=detectors, fknee=fknee, alpha=alpha, NET=NET
    )
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Creating noise model")
    return noise



# Copyright (c) 2015-2018 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

# FIXME:  Restore these unit tests once the relevant code has been updated in
# toast.tod.spt3g.


#
# from ..mpi import MPI
# from .mpi import MPITestCase
#
# import sys
# import os
# import re
# import shutil
#
# import unittest
#
# import numpy as np
# import numpy.testing as nt
#
# from ..dist import *
#
# from .. import qarray as qa
#
# from ..tod import Interval, TODCache, TODGround
#
# from ._helpers import (
#     create_outdir,
#     create_distdata,
#     boresight_focalplane,
#     uniform_chunks,
# )
#
# # This file will only be imported if SPT3G is already available
# from spt3g import core as c3g
# from ..tod import spt3g_utils as s3utils
# from ..tod import spt3g as s3g
#
#
# class Spt3gTest(MPITestCase):
#     def setUp(self):
#         fixture_name = os.path.splitext(os.path.basename(__file__))[0]
#         self.outdir = create_outdir(self.comm, fixture_name)
#
#         self.datawrite = os.path.join(self.outdir, "test_3g")
#         self.dataexport = os.path.join(self.outdir, "export_3g")
#         self.groundexport = os.path.join(self.outdir, "export_3g_ground")
#
#         # Reset the frame file size for these tests, so that we can test
#         # boundary effects.
#         self._original_framefile = s3g.TARGET_FRAMEFILE_SIZE
#
#         # Create observations, divided evenly between groups
#         self.nobs = 4
#         opg = self.nobs
#         if self.comm.size >= 2:
#             opg = self.nobs // 2
#         self.data = create_distdata(self.comm, obs_per_group=opg)
#
#         # Create a set of boresight detectors
#         self.ndet = 4
#         self.rate = 20.0
#
#         self.dnames, self.dquat, depsilon, drate, dnet, dfmin, dfknee, dalpha = boresight_focalplane(
#             self.ndet, samplerate=self.rate
#         )
#
#         # Properties of each observation
#
#         self.obslen = 90.0
#         self.obsgap = 10.0
#
#         self.obstotalsamp = int((self.obslen + self.obsgap) * self.rate)
#         self.obssamp = int(self.obslen * self.rate)
#         self.obsgapsamp = self.obstotalsamp - self.obssamp
#
#         self.obstotal = self.obstotalsamp / self.rate
#         self.obslen = self.obssamp / self.rate
#         self.obsgap = self.obstotal - self.obsgap
#
#         # Properties of the intervals within an observation
#         self.nsub = 12
#         self.subtotsamp = self.obssamp // self.nsub
#         self.subgapsamp = 0
#         self.subsamp = self.subtotsamp - self.subgapsamp
#
#         self.frames = list()
#         for i in range(self.nsub):
#             self.frames.append(self.subtotsamp)
#         if self.obssamp > self.subtotsamp * self.nsub:
#             self.frames.append(self.obssamp - self.subtotsamp * self.nsub)
#
#         # Ground scan properties
#         self.site_lon = "-67:47:10"
#         self.site_lat = "-22:57:30"
#         self.site_alt = 5200.0
#         self.coord = "C"
#         self.azmin = 45
#         self.azmax = 55
#         self.el = 60
#         self.scanrate = 5.0
#         self.scan_accel = 100.0
#         self.CES_start = None
#
#     def tearDown(self):
#         # Reset the frame file size.
#         s3g.TARGET_FRAMEFILE_SIZE = self._original_framefile
#         return
#
#     def meta_setup(self):
#         ret = dict()
#         ret["string"] = "blahblahblah"
#         ret["double"] = -1.234567890123e9
#         ret["int64"] = -100000000000
#         return ret
#
#     def meta_verify(self, dct):
#         nt.assert_equal(dct["string"], "blahblahblah")
#         nt.assert_equal(dct["int64"], -100000000000)
#         nt.assert_almost_equal(dct["double"], -1.234567890123e9)
#         return
#
#     def create_bore(self, total, local):
#         theta_incr = (0.5 * np.pi) / (total - 1)
#         phi_incr = (2.0 * np.pi) / (total - 1)
#
#         theta_start = local[0] * theta_incr
#         phi_start = local[0] * phi_incr
#
#         theta_stop = theta_start + (local[1] - 1) * theta_incr
#         phi_stop = phi_start + (local[1] - 1) * phi_incr
#
#         theta = np.linspace(
#             theta_start, theta_stop, num=local[1], endpoint=True, dtype=np.float64
#         )
#         phi = np.linspace(
#             phi_start, phi_stop, num=local[1], endpoint=True, dtype=np.float64
#         )
#         pa = np.zeros(local[1], dtype=np.float64)
#
#         return qa.from_angles(theta, phi, pa)
#
#     def obs_create(self, comm, name):
#         # fake metadata
#         props = self.meta_setup()
#
#         tod = s3g.TOD3G(
#             comm,
#             comm.size,
#             detectors=self.dquat,
#             samples=self.obssamp,
#             framesizes=self.frames,
#             azel=True,
#             meta=props,
#         )
#
#         obs = dict()
#         obs["name"] = name
#         obs["tod"] = tod
#
#         return obs
#
#     def obs_init(self, obs, start, off):
#         tod = obs["tod"]
#         # Now write the data.  For this test, we simply write the detector
#         # index (as a float) to the detector timestream.  We also flag every
#         # other sample.  For the boresight pointing, we create a fake spiral
#         # pattern.
#
#         detranks, sampranks = tod.grid_size
#         rankdet, ranksamp = tod.grid_ranks
#
#         off = tod.local_samples[0]
#         n = tod.local_samples[1]
#
#         # We use this for both the common and all the detector
#         # flags just to check write/read roundtrip.
#         flags = np.zeros(n, dtype=np.uint8)
#         flags[::2] = 1
#
#         # Everyone writes their timestamps
#
#         incr = 1.0 / self.rate
#         stamps = np.arange(n, dtype=np.float64)
#         stamps *= incr
#         stamps += start + (off * incr)
#         tod.write_times(stamps=stamps)
#
#         # Same with the boresight
#         boresight = self.create_bore(tod.total_samples, tod.local_samples)
#         tod.write_boresight(data=boresight)
#
#         # Just duplicate the RA/DEC quaternions to AZ/EL.  We are just
#         # checking read / write integrity.
#         tod.write_boresight_azel(data=boresight)
#
#         # Fake velocity / position data
#         posvec = np.zeros((n, 3), dtype=np.float64)
#         posvec[:, 2] = 1.0
#         tod.write_velocity(vel=posvec)
#         tod.write_position(pos=posvec)
#
#         # Now the common flags
#         tod.write_common_flags(flags=flags)
#
#         # Detector data
#         fakedata = np.empty(n, dtype=np.float64)
#         for d in tod.local_dets:
#             # get unique detector index and convert to float
#             indx = float(tod.detindx[d])
#             # write data based on this to all local elements
#             fakedata[:] = np.arange(n)
#             fakedata += off
#             fakedata *= indx
#             tod.write(detector=d, data=fakedata)
#             # write detector flags
#             tod.write_flags(detector=d, flags=flags)
#         return
#
#     def obs_zero(self, obs, start, off):
#         tod = obs["tod"]
#         # Write empty data to all fields
#
#         detranks, sampranks = tod.grid_size
#         rankdet, ranksamp = tod.grid_ranks
#
#         off = tod.local_samples[0]
#         n = tod.local_samples[1]
#
#         # We use this for both the common and all the detector
#         # flags just to check write/read roundtrip.
#         flags = np.zeros(n, dtype=np.uint8)
#
#         # Everyone writes their timestamps
#
#         stamps = np.zeros(n, dtype=np.float64)
#         tod.write_times(stamps=stamps)
#
#         # Same with the boresight
#         boresight = np.zeros((n, 4), dtype=np.float64)
#         tod.write_boresight(data=boresight)
#         tod.write_boresight_azel(data=boresight)
#
#         # Fake velocity / position data
#         posvec = np.zeros((n, 3), dtype=np.float64)
#         tod.write_velocity(vel=posvec)
#         tod.write_position(pos=posvec)
#
#         # Now the common flags
#         tod.write_common_flags(flags=flags)
#
#         # Detector data
#         fakedata = np.zeros(n, dtype=np.float64)
#         for d in tod.local_dets:
#             tod.write(detector=d, data=fakedata)
#             tod.write_flags(detector=d, flags=flags)
#         return
#
#     def obs_verify(self, tod, start, off):
#         nlocal = tod.local_samples[1]
#         odd = False
#         if tod.local_samples[0] % 2 != 0:
#             odd = True
#
#         # Verify metadata
#
#         self.meta_verify(tod.meta())
#
#         # Read and compare timestamps
#
#         compstamps = np.arange(nlocal, dtype=np.float64)
#         compstamps /= self.rate
#         compstamps += start + (tod.local_samples[0] / self.rate)
#
#         stamps = tod.read_times()
#         nt.assert_almost_equal(stamps, compstamps)
#         del stamps
#
#         # Read and compare boresight
#
#         compbore = self.create_bore(tod.total_samples, tod.local_samples)
#
#         boresight = tod.read_boresight()
#         nt.assert_almost_equal(boresight, compbore)
#         del boresight
#
#         boresight = tod.read_boresight_azel()
#         nt.assert_almost_equal(boresight, compbore)
#         del boresight
#
#         # flags.  We use this for both the common and all the detector
#         # flags just to check write/read roundtrip.
#
#         compflags = np.zeros(nlocal, dtype=np.uint8)
#         if odd:
#             compflags[1::2] = 1
#         else:
#             compflags[::2] = 1
#
#         flags = tod.read_common_flags()
#         nt.assert_equal(flags, compflags)
#         del flags
#
#         # detector data
#
#         compdata = np.empty(nlocal, dtype=np.float64)
#         for d in tod.local_dets:
#             # get unique detector index and convert to float
#             indx = float(tod.detindx[d])
#             # comparison values
#             compdata[:] = np.arange(tod.local_samples[1])
#             compdata += tod.local_samples[0]
#             compdata *= indx
#             # read and check
#             data = tod.read(detector=d)
#             nt.assert_almost_equal(data, compdata)
#             del data
#
#             # check detector flags
#             flags = tod.read_flags(detector=d)
#             nt.assert_equal(flags, compflags)
#             del flags
#
#         return
#
#     def data_init(self):
#         data = Data(self.data.comm)
#
#         for ob in range(len(self.data.obs)):
#             obsname = "obs_{}_{:02d}".format(self.data.comm.group, ob)
#             obsstart = ob * self.obstotal
#             obsoff = ob * self.obstotalsamp
#             obs = self.obs_create(self.data.comm.comm_group, obsname)
#             self.obs_init(obs, obsstart, obsoff)
#             data.obs.append(obs)
#
#         return data
#
#     def init_ground(self):
#         # Create the simulated TODs
#         for ob in range(len(self.data.obs)):
#             obsname = "obs_{}_{:02d}".format(self.data.comm.group, ob)
#             start = ob * self.obstotal
#             first = ob * self.obstotalsamp
#             tod = TODGround(
#                 self.data.comm.comm_group,
#                 self.dquat,
#                 self.obstotalsamp,
#                 detranks=self.data.comm.group_size,
#                 firsttime=start,
#                 rate=self.rate,
#                 site_lon=self.site_lon,
#                 site_lat=self.site_lat,
#                 site_alt=self.site_alt,
#                 azmin=self.azmin,
#                 azmax=self.azmax,
#                 el=self.el,
#                 coord=self.coord,
#                 scanrate=self.scanrate,
#                 scan_accel=self.scan_accel,
#                 CES_start=self.CES_start,
#             )
#             self.data.obs[ob]["tod"] = tod
#         return
#
#     def data_verify(self, path, prefix):
#         for ob in range(len(self.data.obs)):
#             obsname = "obs_{}_{:02d}".format(self.data.comm.group, ob)
#             obsdir = os.path.join(path, obsname)
#             obsstart = ob * self.obstotal
#             obsoff = ob * self.obstotalsamp
#             tod = s3g.TOD3G(
#                 self.data.comm.comm_group,
#                 self.data.comm.comm_group.size,
#                 path=obsdir,
#                 prefix=prefix,
#             )
#             self.obs_verify(tod, obsstart, obsoff)
#         return
#
#     def test_utils(self):
#         s3g.TARGET_FRAMEFILE_SIZE = 200000
#         # We want to test the frame operations with a process grid that has
#         # multiple ranks in both the sample and detector directions.
#         detranks = self.comm.size
#         if self.comm.size % 2 == 0:
#             detranks = 2
#
#         # Create a simple tod with a cache that we can use for testing.
#         tod_in = TODCache(
#             self.comm,
#             self.dnames,
#             self.obssamp,
#             detquats=self.dquat,
#             detranks=detranks,
#             sampsizes=self.frames,
#         )
#
#         tod_out = TODCache(
#             self.comm,
#             self.dnames,
#             self.obssamp,
#             detquats=self.dquat,
#             detranks=detranks,
#             sampsizes=self.frames,
#         )
#
#         # Write some fake data to this TOD
#         obs_in = {"tod": tod_in}
#         self.obs_init(obs_in, 0.0, 0)
#
#         obs_out = {"tod": tod_out}
#         self.obs_zero(obs_out, 0.0, 0)
#
#         # For timestamps, we need to copy the internal cached timestamps
#         # (float64) into spt3g timestamps.
#         stamps = np.copy(tod_in.cache.reference(tod_in._stamps))
#         stamps *= 1.0e9
#         istamps = stamps.astype(np.int64)
#         tod_in.cache.put("spt3gtime", istamps)
#
#         # Make lists of fields that we are going to write to the spt3g frames.
#         # These use the internal names of the cache objects in a TODCache
#         # class.  We will test that we can dump this data to frames and
#         # restore it.
#
#         common = list()
#         common.append(("spt3gtime", c3g.G3VectorTime, "spt3gtime"))
#         common.append((tod_in._bore, c3g.G3VectorDouble, tod_in._bore))
#         common.append((tod_in._bore_azel, c3g.G3VectorDouble, tod_in._bore_azel))
#         common.append((tod_in._pos, c3g.G3VectorDouble, tod_in._pos))
#         common.append((tod_in._vel, c3g.G3VectorDouble, tod_in._vel))
#         common.append((tod_in._common, c3g.G3VectorUnsignedChar, tod_in._common))
#
#         detfields = [
#             (
#                 "{}{}".format(tod_in._pref_detdata, x),
#                 "{}{}".format(tod_in._pref_detdata, x),
#             )
#             for x in tod_in.detectors
#         ]
#
#         flagfields = [
#             (
#                 "{}{}".format(tod_in._pref_detflags, x),
#                 "{}{}".format(tod_in._pref_detflags, x),
#             )
#             for x in tod_in.detectors
#         ]
#
#         off = 0
#         frames = list()
#         for findx, frm in enumerate(self.frames):
#             # print("cache to frame {} at {}".format(findx, off), flush=True)
#             fdata = s3utils.cache_to_frames(
#                 tod_in,
#                 findx,
#                 1,
#                 [off],
#                 [frm],
#                 common=common,
#                 detector_fields=detfields,
#                 flag_fields=flagfields,
#                 units=c3g.G3TimestreamUnits.Tcmb,
#             )
#             # print("  got ",fdata, flush=True)
#             frames.extend(fdata)
#             off += frm
#
#         # Restore frames to cache
#
#         off = 0
#         for findx, frm in enumerate(self.frames):
#             # print("frame to cache {} at {}".format(findx, off), flush=True)
#             s3utils.frame_to_cache(tod_out, findx, off, frm, frame_data=frames[findx])
#             off += frm
#
#         # Compare input to output
#
#         np.testing.assert_almost_equal(
#             tod_in.cache.reference("spt3gtime"), tod_out.cache.reference("spt3gtime")
#         )
#
#         np.testing.assert_almost_equal(tod_in.read_position(), tod_out.read_position())
#
#         np.testing.assert_almost_equal(tod_in.read_velocity(), tod_out.read_velocity())
#
#         np.testing.assert_almost_equal(
#             tod_in.read_boresight(), tod_out.read_boresight()
#         )
#
#         np.testing.assert_almost_equal(
#             tod_in.read_boresight_azel(), tod_out.read_boresight_azel()
#         )
#
#         np.testing.assert_equal(tod_in.read_common_flags(), tod_out.read_common_flags())
#
#         for det in tod_in.local_dets:
#             np.testing.assert_almost_equal(
#                 tod_in.read(detector=det), tod_out.read(detector=det)
#             )
#             np.testing.assert_equal(
#                 tod_in.read_flags(detector=det), tod_out.read_flags(detector=det)
#             )
#
#         return
#
#     def test_io(self):
#         s3g.TARGET_FRAMEFILE_SIZE = 200000
#         origdata = self.data_init()
#
#         if self.comm.rank == 0:
#             if os.path.isdir(self.datawrite):
#                 shutil.rmtree(self.datawrite)
#         self.comm.barrier()
#
#         dumper = s3g.Op3GExport(
#             self.datawrite,
#             s3g.TOD3G,
#             use_todchunks=True,
#             export_opts={"prefix": "test"},
#         )
#         dumper.exec(origdata)
#
#         # print("{}: Done with export 1".format(self.comm.rank), flush=True)
#         self.comm.barrier()
#
#         self.data_verify(self.datawrite, "test")
#
#         # print("{}: Done with verify 1".format(self.comm.rank), flush=True)
#         self.comm.barrier()
#
#         loaddata = s3g.load_spt3g(
#             self.data.comm,
#             self.data.comm.comm_group.size,
#             self.datawrite,
#             "test",
#             s3g.obsweight_spt3g,
#             s3g.TOD3G,
#         )
#
#         # print("{}: Done with load".format(self.comm.rank), flush=True)
#         self.comm.barrier()
#
#         if self.comm.rank == 0:
#             if os.path.isdir(self.dataexport):
#                 shutil.rmtree(self.dataexport)
#         self.comm.barrier()
#
#         exporter = s3g.Op3GExport(
#             self.dataexport,
#             s3g.TOD3G,
#             use_todchunks=True,
#             export_opts={"prefix": "test"},
#         )
#         exporter.exec(loaddata)
#
#         # print("{}: Done with export 2".format(self.comm.rank), flush=True)
#         self.comm.barrier()
#
#         self.data_verify(self.dataexport, "test")
#
#         # print("{}: Done with verify 2".format(self.comm.rank), flush=True)
#         self.comm.barrier()
#
#         return
#
#     def test_ground(self):
#         s3g.TARGET_FRAMEFILE_SIZE = 400000
#         # Create simulated tods in memory
#         self.init_ground()
#
#         # Export this
#         if self.comm.rank == 0:
#             if os.path.isdir(self.groundexport):
#                 shutil.rmtree(self.groundexport)
#         self.comm.barrier()
#
#         dumper = s3g.Op3GExport(
#             self.groundexport,
#             s3g.TOD3G,
#             use_todchunks=True,
#             export_opts={"prefix": "test"},
#         )
#         dumper.exec(self.data)
#
#         # Load it back in.
#
#         loaddata = s3g.load_spt3g(
#             self.data.comm,
#             self.data.comm.comm_group.size,
#             self.groundexport,
#             "test",
#             s3g.obsweight_spt3g,
#             s3g.TOD3G,
#         )
#
#         return



#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""
Simpler version of the ground simulation script
"""

import argparse
import dateutil.parser
import os
import pickle
import sys
import traceback

import numpy as np

from toast.mpi import get_world, Comm

from toast.dist import distribute_uniform, Data

from toast.utils import Logger, Environment, memreport

from toast.timing import function_timer, GlobalTimers, Timer, gather_timers
from toast.timing import dump as dump_timing

import toast.qarray as qa
from toast.tod import OpCacheCopy, plot_focalplane, OpCacheClear
from toast.todmap import TODGround

from toast.pipeline_tools import (
    add_dist_args,
    add_debug_args,
    get_time_communicators,
    get_comm,
    add_polyfilter_args,
    apply_polyfilter,
    add_groundfilter_args,
    apply_groundfilter,
    add_atmosphere_args,
    add_noise_args,
    simulate_noise,
    add_gainscrambler_args,
    scramble_gains,
    add_pointing_args,
    expand_pointing,
    add_madam_args,
    setup_madam,
    apply_madam,
    add_sky_map_args,
    add_pysm_args,
    scan_sky_signal,
    simulate_sky_signal,
    copy_signal,
    add_tidas_args,
    output_tidas,
    add_spt3g_args,
    output_spt3g,
    add_todground_args,
    get_breaks,
    Focalplane,
    load_schedule,
    add_mc_args,
    add_binner_args,
    init_binner,
    apply_binner,
)


def parse_arguments(comm):
    timer = Timer()
    log = Logger.get()

    parser = argparse.ArgumentParser(
        description="Simulate ground-based boresight pointing.  Simulate "
        "and map astrophysical signal.",
        fromfile_prefix_chars="@",
    )

    add_dist_args(parser)
    add_debug_args(parser)
    add_todground_args(parser)
    add_pointing_args(parser)
    add_polyfilter_args(parser)
    add_groundfilter_args(parser)
    add_gainscrambler_args(parser)
    add_noise_args(parser)
    add_sky_map_args(parser)
    add_tidas_args(parser)

    parser.add_argument(
        "--outdir", required=False, default="out", help="Output directory"
    )

    add_madam_args(parser)
    add_binner_args(parser)

    parser.add_argument(
        "--madam",
        required=False,
        action="store_true",
        help="Use libmadam for map-making",
        dest="use_madam",
    )
    parser.add_argument(
        "--no-madam",
        required=False,
        action="store_false",
        help="Do not use libmadam for map-making [default]",
        dest="use_madam",
    )
    parser.set_defaults(use_madam=False)

    parser.add_argument(
        "--focalplane",
        required=False,
        default=None,
        help="Pickle file containing a dictionary of detector "
        "properties.  The keys of this dict are the detector "
        "names, and each value is also a dictionary with keys "
        '"quat" (4 element ndarray), "fwhm" (float, arcmin), '
        '"fknee" (float, Hz), "alpha" (float), and '
        '"NET" (float).  For optional plotting, the key "color"'
        " can specify a valid matplotlib color string.",
    )

    try:
        args = parser.parse_args()
    except SystemExit:
        sys.exit(0)

    if args.tidas is not None:
        if not tidas_available:
            raise RuntimeError("TIDAS not found- cannot export")

    if comm.comm_world is None or comm.world_rank == 0:
        log.info("All parameters:")
        for ag in vars(args):
            log.info("{} = {}".format(ag, getattr(args, ag)))

    if args.group_size:
        comm = Comm(groupsize=args.group_size)

    if comm.comm_world is None or comm.comm_world.rank == 0:
        os.makedirs(args.outdir, exist_ok=True)

    if comm.comm_world is None or comm.world_rank == 0:
        timer.report_clear("Parsed parameters")

    return args, comm


def load_focalplane(args, comm, schedule):
    focalplane = None

    # Load focalplane information

    if comm.comm_world is None or comm.comm_world.rank == 0:
        if args.focalplane is None:
            detector_data = {}
            ZAXIS = np.array([0, 0, 1.0])
            # in this case, create a fake detector at the boresight
            # with a pure white noise spectrum.
            fake = {}
            fake["quat"] = np.array([0, 0, 0, 1.0])
            fake["fwhm"] = 30.0
            fake["fknee"] = 0.0
            fake["fmin"] = 1e-9
            fake["alpha"] = 1.0
            fake["NET"] = 1.0
            fake["color"] = "r"
            detector_data["bore1"] = fake
            # Second detector at 22.5 degree polarization angle
            fake2 = {}
            zrot = qa.rotation(ZAXIS, np.radians(22.5))
            fake2["quat"] = qa.mult(fake["quat"], zrot)
            fake2["fwhm"] = 30.0
            fake2["fknee"] = 0.0
            fake2["fmin"] = 1e-9
            fake2["alpha"] = 1.0
            fake2["NET"] = 1.0
            fake2["color"] = "r"
            detector_data["bore2"] = fake2
            # Third detector at 45 degree polarization angle
            fake3 = {}
            zrot = qa.rotation(ZAXIS, np.radians(45))
            fake3["quat"] = qa.mult(fake["quat"], zrot)
            fake3["fwhm"] = 30.0
            fake3["fknee"] = 0.0
            fake3["fmin"] = 1e-9
            fake3["alpha"] = 1.0
            fake3["NET"] = 1.0
            fake3["color"] = "r"
            detector_data["bore3"] = fake3
            # Fourth detector at 67.5 degree polarization angle
            fake4 = {}
            zrot = qa.rotation(ZAXIS, np.radians(67.5))
            fake4["quat"] = qa.mult(fake["quat"], zrot)
            fake4["fwhm"] = 30.0
            fake4["fknee"] = 0.0
            fake4["fmin"] = 1e-9
            fake4["alpha"] = 1.0
            fake4["NET"] = 1.0
            fake4["color"] = "r"
            detector_data["bore4"] = fake4
            focalplane = Focalplane(
                detector_data=detector_data, sample_rate=args.sample_rate
            )
        else:
            focalplane = Focalplane(
                fname_pickle=args.focalplane, sample_rate=args.sample_rate
            )
    if comm.comm_world is not None:
        focalplane = comm.comm_world.bcast(focalplane, root=0)

    if args.debug:
        if comm.comm_world is None or comm.comm_world.rank == 0:
            outfile = "{}/focalplane.png".format(args.outdir)
            plot_focalplane(focalplane, 6, 6, outfile)

    schedule.telescope.focalplane = focalplane
    detweights = focalplane.detweights

    return detweights


def create_observations(args, comm, schedule):
    """Simulate constant elevation scans.

    Simulate constant elevation scans at "site" matching entries in
    "all_ces".  Each operational day is assigned to a different
    process group to allow making day maps.

    """
    timer = Timer()
    log = Logger.get()

    data = Data(comm)

    telescope = schedule.telescope
    site = telescope.site
    focalplane = telescope.focalplane
    all_ces = schedule.ceslist
    nces = len(all_ces)

    breaks = get_breaks(comm, all_ces, args)

    nbreak = len(breaks)

    groupdist = distribute_uniform(nces, comm.ngroups, breaks=breaks)
    group_firstobs = groupdist[comm.group][0]
    group_numobs = groupdist[comm.group][1]

    if comm.comm_group is not None:
        ndetrank = comm.comm_group.size
    else:
        ndetrank = 1

    for ices in range(group_firstobs, group_firstobs + group_numobs):
        ces = all_ces[ices]
        totsamples = int((ces.stop_time - ces.start_time) * args.sample_rate)

        # create the single TOD for this observation

        try:
            tod = TODGround(
                comm.comm_group,
                focalplane.detquats,
                totsamples,
                detranks=ndetrank,
                firsttime=ces.start_time,
                rate=args.sample_rate,
                site_lon=site.lon,
                site_lat=site.lat,
                site_alt=site.alt,
                azmin=ces.azmin,
                azmax=ces.azmax,
                el=ces.el,
                scanrate=args.scan_rate,
                scan_accel=args.scan_accel,
                cosecant_modulation=args.scan_cosecant_modulate,
                CES_start=None,
                CES_stop=None,
                sun_angle_min=args.sun_angle_min,
                coord=args.coord,
                sampsizes=None,
                report_timing=args.debug,
            )
        except RuntimeError as e:
            raise RuntimeError("Failed to create the CES scan: {}".format(e))

        # Create the (single) observation

        ob = {}
        ob["name"] = "CES-{}-{}-{}".format(ces.name, ces.scan, ces.subscan)
        ob["tod"] = tod
        if len(tod.subscans) > 0:
            ob["intervals"] = tod.subscans
        else:
            raise RuntimeError("{} has no valid intervals".format(ob["name"]))
        ob["baselines"] = None
        ob["noise"] = focalplane.noise
        ob["id"] = int(ces.mjdstart * 10000)

        data.obs.append(ob)

    for ob in data.obs:
        tod = ob["tod"]
        tod.free_azel_quats()

    if comm.comm_world is None or comm.comm_group.rank == 0:
        log.info("Group # {:4} has {} observations.".format(comm.group, len(data.obs)))

    if len(data.obs) == 0:
        raise RuntimeError(
            "Too many tasks. Every MPI task must "
            "be assigned to at least one observation."
        )

    if comm.world_rank == 0:
        timer.report_clear("Simulate scans")

    return data


def setup_sigcopy(args, comm, signalname):
    """Setup for copying the signal so we can run filter+bin and Madam."""
    if args.use_madam:
        signalname_madam = "signal_madam"
        sigcopy_madam = OpCacheCopy(signalname, signalname_madam)
        sigclear = OpCacheClear(signalname)
    else:
        signalname_madam = signalname
        sigcopy_madam = None
        sigclear = None

    return signalname_madam, sigcopy_madam, sigclear


def setup_output(args, comm):
    outpath = "{}".format(args.outdir)
    if comm.world_rank == 0:
        if not os.path.isdir(outpath):
            try:
                os.makedirs(outpath)
            except FileExistsError:
                pass
    return outpath


def copy_signal_madam(args, comm, data, sigcopy_madam):
    """Make a copy of the TOD for Madam."""
    if sigcopy_madam is not None:
        if comm.world_rank == 0:
            print("Making a copy of the TOD for Madam", flush=args.flush)
        sigcopy_madam.exec(data)

    return


def clear_signal(args, comm, data, sigclear):
    if sigclear is not None:
        if comm.world_rank == 0:
            print("Clearing filtered signal")
        sigclear.exec(data)
    return


def main():
    log = Logger.get()
    gt = GlobalTimers.get()
    gt.start("toast_ground_sim (total)")

    mpiworld, procs, rank, comm = get_comm()

    args, comm = parse_arguments(comm)

    # Initialize madam parameters

    madampars = setup_madam(args)

    # Load and broadcast the schedule file

    schedule = load_schedule(args, comm)[0]

    # load or simulate the focalplane

    detweights = load_focalplane(args, comm, schedule)

    # Create the TOAST data object to match the schedule.  This will
    # include simulating the boresight pointing.

    data = create_observations(args, comm, schedule)

    # Expand boresight quaternions into detector pointing weights and
    # pixel numbers

    expand_pointing(args, comm, data)

    # Scan input map

    signalname = scan_sky_signal(args, comm, data, "signal")

    # Simulate noise

    if signalname is None:
        signalname = "signal"
        mc = 0
        simulate_noise(args, comm, data, mc, signalname)

    # Set up objects to take copies of the TOD at appropriate times

    signalname_madam, sigcopy_madam, sigclear = setup_sigcopy(args, comm, signalname)

    npp, zmap = init_binner(args, comm, data, detweights)

    output_tidas(args, comm, data, signalname)

    outpath = setup_output(args, comm)

    # Make a copy of the signal for Madam

    copy_signal_madam(args, comm, data, sigcopy_madam)

    # Bin unprocessed signal for reference

    apply_binner(args, comm, data, npp, zmap, detweights, outpath, signalname)

    if args.apply_polyfilter or args.apply_groundfilter:

        # Filter signal

        apply_polyfilter(args, comm, data, signalname)

        apply_groundfilter(args, comm, data, signalname)

        # Bin the filtered signal

        apply_binner(
            args,
            comm,
            data,
            npp,
            zmap,
            detweights,
            outpath,
            signalname,
            prefix="filtered",
        )

    data.obs[0]["tod"].cache.report()

    clear_signal(args, comm, data, sigclear)

    data.obs[0]["tod"].cache.report()

    # Now run Madam on the unprocessed copy of the signal

    if args.use_madam:
        apply_madam(args, comm, data, madampars, outpath, detweights, signalname_madam)

    gt.stop_all()
    if mpiworld is not None:
        mpiworld.barrier()
    timer = Timer()
    timer.start()
    alltimers = gather_timers(comm=mpiworld)
    if comm.world_rank == 0:
        out = os.path.join(args.outdir, "timing")
        dump_timing(alltimers, out)
        timer.report_clear("Gather and dump timing info")
    return


if __name__ == "__main__":
    try:
        main()
    except Exception:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        if procs == 1:
            raise
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None:
            mpiworld.Abort(6)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os
import numpy as np

from ..fft import r1d_forward, r1d_backward

from ..rng import random

from ._helpers import create_outdir


class FFTTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)
        self.length = 65536
        self.input_one = random(self.length, counter=[0, 0], key=[0, 0])
        self.compare_one = np.copy(self.input_one)
        self.nbatch = 5
        self.input_batch = np.zeros((self.nbatch, self.length), dtype=np.float64)
        for b in range(self.nbatch):
            self.input_batch[b, :] = random(self.length, counter=[0, 0], key=[0, b])
        self.compare_batch = np.array(self.input_batch)

    def test_roundtrip(self):
        output = r1d_forward(self.input_one)
        check = r1d_backward(output)

        if (self.comm is None) or (self.comm.rank == 0):
            # One process dumps debugging info
            import matplotlib.pyplot as plt

            savefile = os.path.join(self.outdir, "out_one_fft.txt")
            np.savetxt(savefile, np.transpose([self.compare_one, check]), delimiter=" ")
            fig = plt.figure(figsize=(12, 8), dpi=72)
            ax = fig.add_subplot(1, 1, 1, aspect="auto")
            ax.plot(np.arange(self.length), check, c="red", label="Output")
            ax.plot(np.arange(self.length), self.compare_one, c="black", label="Input")
            ax.legend(loc=1)
            plt.title("FFT One Input and Output")
            savefile = os.path.join(self.outdir, "out_one_fft.png")
            plt.savefig(savefile)
            plt.close()

        np.testing.assert_array_almost_equal(check, self.compare_one)

        output = r1d_forward(self.input_batch)
        check = r1d_backward(output)

        if (self.comm is None) or (self.comm.rank == 0):
            # One process dumps debugging info
            import matplotlib.pyplot as plt

            for b in range(self.nbatch):
                savefile = os.path.join(self.outdir, "out_batch_{}_fft.txt".format(b))
                np.savetxt(
                    savefile,
                    np.transpose([self.compare_batch[b], check[b]]),
                    delimiter=" ",
                )
                fig = plt.figure(figsize=(12, 8), dpi=72)
                ax = fig.add_subplot(1, 1, 1, aspect="auto")
                ax.plot(np.arange(self.length), check[b], c="red", label="Output")
                ax.plot(
                    np.arange(self.length),
                    self.compare_batch[b],
                    c="black",
                    label="Input",
                )
                ax.legend(loc=1)
                plt.title("FFT Batch {} Input and Output".format(b))
                savefile = os.path.join(self.outdir, "out_batch_{}_fft.png".format(b))
                plt.savefig(savefile)
                plt.close()

        np.testing.assert_array_almost_equal(check, self.compare_batch)
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

from ..timing import function_timer, Timer

from ..mpi import MPI

import healpy as hp

from ..op import Operator

from ..cache import Cache

from .._libtoast import global_to_local as libtoast_global_to_local


class DistPixels(object):
    """A distributed map with multiple values per pixel.

    Pixel domain data is distributed across an MPI communicator.  each
    process has a local data stored in one or more "submaps".  The size
    of the submap can be tuned to balance storage (smaller submap size
    means fewer wasted pixels stored) and ease of indexing (larger
    submap means faster global-to-local pixel lookups).

    Although multiple processes may have the same submap of data stored
    locally, the lowest-rank process that has a given submap is the
    "owner" for operations like serialization.

    Args:
        data (toast.Data) : TOAST data object containing the
            pixelization metadata
        comm (mpi4py.MPI.Comm): the MPI communicator containing all
            processes (if None, use data.comm.comm_world).
        npix (int): the total number of pixels.
        npix_submap (int): the locally stored data is in units of this size.
        local_submaps (array): the list of local submaps (integers).
        nnz (int): the number of values per pixel.
        nest (bool): nested pixel order flag
        pixels (str):  cache prefix used for pixel numbers
    """

    def __init__(
        self,
        data,
        comm=None,
        nnz=1,
        dtype=np.float64,
        npix=None,  # if data is None
        npix_submap=None,  # if data is None
        local_submaps=None,  # if data is None
        nest=True,
        pixels="pixels",
    ):
        if data is None:
            self._npix = npix
            self._npix_submap = npix_submap
            self._local_submaps = local_submaps
            self._comm = comm
            if self._npix % self._npix_submap != 0:
                raise RuntimeError(
                    "submap size must evenly divide into total number of pixels"
                )
            self._nglob = self._npix // self._npix_submap
        else:
            self._npix = data["{}_npix".format(pixels)]
            self._npix_submap = data["{}_npix_submap".format(pixels)]
            self._local_submaps = data["{}_local_submaps".format(pixels)]
            self._nglob = data["{}_nsubmap".format(pixels)]
            if comm is None:
                self._comm = data.comm.comm_world
            else:
                self._comm = comm
        self._nnz = nnz
        self._dtype = dtype
        self._nest = nest

        self._glob2loc = None
        self._cache = Cache()
        self._commsize = 5000000

        # our data is a 3D array of submap, pixel, values
        # we allocate this as a contiguous block

        self.data = None
        self.flatdata = None
        if self._local_submaps is None:
            self._nsub = 0
        else:
            if len(self._local_submaps) == 0:
                self._nsub = 0
            else:
                self._nsub = len(self._local_submaps)
                self._glob2loc = self._cache.create(
                    "glob2loc", np.int64, (self._nglob,)
                )
                self._glob2loc[:] = -1
                for ilocal_submap, iglobal_submap in enumerate(self._local_submaps):
                    self._glob2loc[iglobal_submap] = ilocal_submap
                if (self._npix_submap * self._local_submaps.max()) > self._npix:
                    raise RuntimeError("local submap indices out of range")
                self.data = self._cache.create(
                    "data", dtype, (self._nsub, self._npix_submap, self._nnz)
                )
                self.flatdata = self.data.view()
                self.flatdata.shape = tuple(
                    [self._nsub * self._npix_submap * self._nnz]
                )

    def __del__(self):
        if self._glob2loc is not None:
            del self._glob2loc
        if self.data is not None:
            del self.data
        self._cache.clear()

    @property
    def comm(self):
        """(mpi4py.MPI.Comm): The MPI communicator used (or None)"""
        return self._comm

    @property
    def npix(self):
        """(int): The global number of pixels."""
        return self._npix

    @property
    def nnz(self):
        """(int): The number of non-zero values per pixel."""
        return self._nnz

    @property
    def dtype(self):
        """(numpy.dtype): The data type of the values."""
        return self._dtype

    @property
    def local_submaps(self):
        """(array): The list of local submaps or None if process has no data."""
        return self._local_submaps

    @property
    def npix_submap(self):
        """(int): The number of pixels in each submap."""
        return self._npix_submap

    @property
    def nsubmap(self):
        """(int): The number of submaps stored on this process."""
        return self._nsub

    @property
    def nested(self):
        """(bool): If True, data is HEALPix NESTED ordering."""
        return self._nest

    @function_timer
    def global_to_local(self, gl):
        """Convert global pixel indices into the local submap and pixel.

        Args:
            gl (int): The global pixel number.

        Returns:
            (tuple):  A tuple containing the local submap index (int) and the
                pixel index local to that submap (int).

        """
        return libtoast_global_to_local(gl, self._npix_submap, self._glob2loc)

    @function_timer
    def duplicate(self, copy=True, nnz=None):
        """Perform a deep copy of the distributed data.

        Returns:
            (DistPixels): A copy of the object.

        """
        ret = DistPixels(
            None,
            comm=self._comm,
            npix=self._npix,
            npix_submap=self._npix_submap,
            local_submaps=self._local_submaps,
            nnz=(self._nnz if nnz is None else nnz),
            dtype=self._dtype,
        )
        if self.data is not None and copy:
            ret.data[:, :, :] = self.data
        return ret

    def _comm_nsubmap(self, bytes):
        """Given a buffer size, compute the number of submaps to communicate.

        Args:
            bytes (int):  The number of bytes.

        Returns:
            (int):  The number of submaps in each buffer.

        """
        dbytes = self._dtype(1).itemsize
        nsub = int(bytes / (dbytes * self._npix_submap * self._nnz))
        if nsub == 0:
            nsub = 1
        allsub = int(self._npix / self._npix_submap)
        if nsub > allsub:
            nsub = allsub
        return nsub

    @function_timer
    def allreduce(self, comm_bytes=None):
        """Perform a buffered allreduce of the pixel domain data.

        Args:
            comm_bytes (int): The approximate message size to use.

        Returns:
            None.

        """
        if self._comm is None:
            return

        if comm_bytes is None:
            comm_bytes = self._commsize
        comm_submap = self._comm_nsubmap(comm_bytes)
        nsub = int(self._npix / self._npix_submap)

        sendbuf = np.zeros(
            comm_submap * self._npix_submap * self._nnz, dtype=self._dtype
        )
        sendview = sendbuf.reshape(comm_submap, self._npix_submap, self._nnz)

        recvbuf = np.zeros(
            comm_submap * self._npix_submap * self._nnz, dtype=self._dtype
        )
        recvview = recvbuf.reshape(comm_submap, self._npix_submap, self._nnz)

        owners = np.zeros(nsub, dtype=np.int32)
        owners.fill(self._comm.size)
        for m in self._local_submaps:
            owners[m] = self._comm.rank
        allowners = np.zeros_like(owners)
        self._comm.Allreduce(owners, allowners, op=MPI.MIN)

        submap_off = 0
        ncomm = comm_submap

        while submap_off < nsub:
            if submap_off + ncomm > nsub:
                ncomm = nsub - submap_off
            if (
                np.sum(allowners[submap_off : submap_off + ncomm])
                != ncomm * self._comm.size
            ):
                # At least one submap has some hits.  Do the allreduce.
                # Otherwise we would skip this buffer to avoid reducing a
                # bunch of zeros.
                for c in range(ncomm):
                    glob = submap_off + c
                    if glob in self._local_submaps:
                        # copy our data in.
                        loc = self._glob2loc[glob]
                        sendview[c, :, :] = self.data[loc, :, :]

                self._comm.Allreduce(sendbuf, recvbuf, op=MPI.SUM)

                for c in range(ncomm):
                    glob = submap_off + c
                    if glob in self._local_submaps:
                        # copy the reduced data
                        loc = self._glob2loc[glob]
                        self.data[loc, :, :] = recvview[c, :, :]

                sendbuf.fill(0)
                recvbuf.fill(0)

            submap_off += ncomm

        return

    @function_timer
    def read_healpix_fits(self, path, comm_bytes=None):
        """Read and broadcast a HEALPix FITS table.

        The root process opens the FITS file in memmap mode and iterates over
        chunks of the map in a way to minimize cache misses in the internal
        FITS buffer.  Chunks of submaps are broadcast to all processes, and
        each process copies data to its local submaps.

        Args:
            path (str): The path to the FITS file.
            comm_bytes (int): The approximate message size to use.

        Returns:
            None

        """
        rank = 0
        if self._comm is not None:
            rank = self._comm.rank

        if comm_bytes is None:
            comm_bytes = self._commsize
        comm_submap = self._comm_nsubmap(comm_bytes)

        # we make the assumption that FITS binary tables are still stored in
        # blocks of 2880 bytes just like always...
        dbytes = self._dtype(1).itemsize
        rowbytes = self._nnz * dbytes
        optrows = int(2880 / rowbytes)

        # get a tuple of all columns in the table.  We choose memmap here so
        # that we can (hopefully) read through all columns in chunks such that
        # we only ever have a couple FITS blocks in memory.
        fdata = None
        if rank == 0:
            # Check that the file is in expected format
            errors = ""
            h = hp.fitsfunc.pf.open(path, "readonly")
            nside = hp.npix2nside(self._npix)
            nside_map = h[1].header["nside"]
            if nside_map != nside:
                errors += "Wrong NSide: {} has {}, expected {}\n" "".format(
                    path, nside_map, nside
                )
            map_nested = False
            if "order" in h[1].header and "NEST" in h[1].header["order"].upper():
                map_nested = True
            if "ordering" in h[1].header and "NEST" in h[1].header["ordering"].upper():
                map_nested = True
            if map_nested != self._nest:
                errors += (
                    "Wrong ordering: {} has nest={}, expected nest={}\n"
                    "".format(path, map_nested, self._nest)
                )
            map_nnz = h[1].header["tfields"]
            if map_nnz != self._nnz:
                errors += "Wrong number of columns: {} has {}, expected {}\n" "".format(
                    path, map_nnz, self._nnz
                )
            h.close()
            if len(errors) != 0:
                raise RuntimeError(errors)
            # Now read the map
            fdata = hp.read_map(
                path,
                field=None,
                dtype=self._dtype,
                memmap=True,
                nest=self._nest,
                verbose=False,
            )
            if self._nnz == 1:
                fdata = (fdata,)

        buf = np.zeros(comm_submap * self._npix_submap * self._nnz, dtype=self._dtype)
        view = buf.reshape(comm_submap, self._npix_submap, self._nnz)

        in_off = 0
        out_off = 0
        submap_off = 0

        rows = optrows
        while in_off < self._npix:
            if in_off + rows > self._npix:
                rows = self._npix - in_off
            # is this the last block for this communication?
            islast = False
            copyrows = rows
            if out_off + rows > (comm_submap * self._npix_submap):
                copyrows = (comm_submap * self._npix_submap) - out_off
                islast = True

            if rank == 0:
                for col in range(self._nnz):
                    coloff = (out_off * self._nnz) + col
                    buf[coloff : coloff + (copyrows * self._nnz) : self._nnz] = fdata[
                        col
                    ][in_off : in_off + copyrows]

            out_off += copyrows
            in_off += copyrows

            if islast:
                if self._comm is not None:
                    self._comm.Bcast(buf, root=0)
                # loop over these submaps, and copy any that we are assigned
                for sm in range(submap_off, submap_off + comm_submap):
                    if sm in self._local_submaps:
                        loc = self._glob2loc[sm]
                        self.data[loc, :, :] = view[sm - submap_off, :, :]
                out_off = 0
                submap_off += comm_submap
                buf.fill(0)
                islast = False

        # flush the remaining buffer
        if out_off > 0:
            if self._comm is not None:
                self._comm.Bcast(buf, root=0)
            # loop over these submaps, and copy any that we are assigned
            for sm in range(submap_off, submap_off + comm_submap):
                if sm in self._local_submaps:
                    loc = self._glob2loc[sm]
                    self.data[loc, :, :] = view[sm - submap_off, :, :]
        return

    @function_timer
    def broadcast_healpix_map(self, fdata, comm_bytes=None):
        """Distribute a map located on a single process.

        The root process takes a map in memory and distributes it.   Chunks of submaps
        are broadcast to all processes, and each process copies data to its local
        submaps.

        Args:
            fdata (array): The input data (only significant on process 0).
            comm_bytes (int): The approximate message size to use.

        Returns:
            None

        """
        rank = 0
        if self._comm is not None:
            rank = self._comm.rank
        if comm_bytes is None:
            comm_bytes = self._commsize
        comm_submap = self._comm_nsubmap(comm_bytes)

        # we make the assumption that FITS binary tables are still stored in
        # blocks of 2880 bytes just like always...
        dbytes = self._dtype(1).itemsize
        rowbytes = self._nnz * dbytes
        optrows = int(2880 / rowbytes)

        # get a tuple of all columns in the table.  We choose memmap here so
        # that we can (hopefully) read through all columns in chunks such that
        # we only ever have a couple FITS blocks in memory.
        if rank == 0:
            if self._nnz == 1:
                fdata = (fdata,)

        buf = np.zeros(comm_submap * self._npix_submap * self._nnz, dtype=self._dtype)
        view = buf.reshape(comm_submap, self._npix_submap, self._nnz)

        in_off = 0
        out_off = 0
        submap_off = 0

        rows = optrows
        while in_off < self._npix:
            if in_off + rows > self._npix:
                rows = self._npix - in_off
            # is this the last block for this communication?
            islast = False
            copyrows = rows
            if out_off + rows > (comm_submap * self._npix_submap):
                copyrows = (comm_submap * self._npix_submap) - out_off
                islast = True

            if rank == 0:
                for col in range(self._nnz):
                    coloff = (out_off * self._nnz) + col
                    buf[coloff : coloff + (copyrows * self._nnz) : self._nnz] = fdata[
                        col
                    ][in_off : in_off + copyrows]

            out_off += copyrows
            in_off += copyrows

            if islast:
                if self._comm is not None:
                    self._comm.Bcast(buf, root=0)
                # loop over these submaps, and copy any that we are assigned
                for sm in range(submap_off, submap_off + comm_submap):
                    if sm in self._local_submaps:
                        loc = self._glob2loc[sm]
                        self.data[loc, :, :] = view[sm - submap_off, :, :]
                out_off = 0
                submap_off += comm_submap
                buf.fill(0)
                islast = False

        # flush the remaining buffer

        if out_off > 0:
            if self._comm is not None:
                self._comm.Bcast(buf, root=0)
            # loop over these submaps, and copy any that we are assigned
            for sm in range(submap_off, submap_off + comm_submap):
                if sm in self._local_submaps:
                    loc = self._glob2loc[sm]
                    self.data[loc, :, :] = view[sm - submap_off, :, :]
        return

    @function_timer
    def write_healpix_fits(self, path, comm_bytes=None):
        """Write data to a HEALPix format FITS table.

        The data across all processes is assumed to be synchronized (the
        data for a given submap shared between processes is identical).  The
        lowest rank process sharing each submap sends their copy to the root
        process for writing.

        Args:
            path (str): The path to the FITS file.
            comm_bytes (int): The approximate message size to use.

        """
        rank = 0
        if self._comm is not None:
            rank = self._comm.rank

        if comm_bytes is None:
            comm_bytes = self._commsize

        # We will reduce some number of whole submaps at a time.
        # Find the number of submaps that fit into the requested
        # communication size.
        dbytes = self._dtype(1).itemsize
        comm_submap = int(comm_bytes / (dbytes * self._npix_submap * self._nnz))
        if comm_submap == 0:
            comm_submap = 1

        nsubmap = int(self._npix / self._npix_submap)
        if nsubmap * self._npix_submap < self._npix:
            nsubmap += 1

        # Determine which processes "own" each submap.

        if self._comm is None:
            NO_OWNER = 1
            allowners = np.zeros(nsubmap, dtype=np.int32)
            allowners.fill(NO_OWNER)
            for m in self._local_submaps:
                allowners[m] = rank
        else:
            NO_OWNER = self._comm.size
            owners = np.zeros(nsubmap, dtype=np.int32)
            owners.fill(NO_OWNER)
            for m in self._local_submaps:
                owners[m] = self._comm.rank
            allowners = np.zeros_like(owners)
            self._comm.Allreduce(owners, allowners, op=MPI.MIN)

        # this function requires lots of RAM, since it accumulates the
        # full map on one process before writing.

        # use a cache to store the local map, so that we can be sure to
        # free the memory afterwards

        fdata = None
        temp = None
        if rank == 0:
            fdata = []
            temp = Cache()
            for col in range(self._nnz):
                name = "col{}".format(col)
                temp.create(name, self._dtype, (self._npix,))
                fdata.append(temp.reference(name))

        if self._comm is None:
            dbuf = np.zeros(
                comm_submap * self._npix_submap * self._nnz, dtype=self._dtype
            )
            dview = dbuf.reshape(comm_submap, self._npix_submap, self._nnz)

            submap_off = 0
            ncomm = comm_submap
            while submap_off < nsubmap:
                if submap_off + ncomm > nsubmap:
                    ncomm = nsubmap - submap_off
                if np.any(allowners[submap_off : submap_off + ncomm] != NO_OWNER):
                    # at least one submap has some hits
                    for c in range(ncomm):
                        if allowners[submap_off + c] == NO_OWNER:
                            dview[c, :, :] = 0
                        else:
                            dview[c, :, :] = self.data[
                                self._glob2loc[submap_off + c], :, :
                            ]
                    # copy into FITS buffers
                    for c in range(ncomm):
                        sampoff = (submap_off + c) * self._npix_submap
                        for col in range(self._nnz):
                            fdata[col][sampoff : sampoff + self._npix_submap] = dview[
                                c, :, col
                            ]
                submap_off += ncomm
        else:
            sendbuf = np.zeros(
                comm_submap * self._npix_submap * self._nnz, dtype=self._dtype
            )
            sendview = sendbuf.reshape(comm_submap, self._npix_submap, self._nnz)

            recvbuf = None
            recvview = None
            if rank == 0:
                recvbuf = np.zeros(
                    comm_submap * self._npix_submap * self._nnz, dtype=self._dtype
                )
                recvview = recvbuf.reshape(comm_submap, self._npix_submap, self._nnz)

            submap_off = 0
            ncomm = comm_submap
            while submap_off < nsubmap:
                if submap_off + ncomm > nsubmap:
                    ncomm = nsubmap - submap_off
                if np.any(allowners[submap_off : submap_off + ncomm] != NO_OWNER):
                    # at least one submap has some hits.  reduce.
                    for c in range(ncomm):
                        if allowners[submap_off + c] == self._comm.rank:
                            sendview[c, :, :] = self.data[
                                self._glob2loc[submap_off + c], :, :
                            ]
                    self._comm.Reduce(sendbuf, recvbuf, op=MPI.SUM, root=0)
                    if rank == 0:
                        # copy into FITS buffers
                        for c in range(ncomm):
                            sampoff = (submap_off + c) * self._npix_submap
                            for col in range(self._nnz):
                                fdata[col][
                                    sampoff : sampoff + self._npix_submap
                                ] = recvview[c, :, col]
                    sendbuf.fill(0)
                    if rank == 0:
                        recvbuf.fill(0)
                submap_off += ncomm

        if rank == 0:
            if os.path.isfile(path):
                os.remove(path)
            hp.write_map(
                path, fdata, dtype=self._dtype, fits_IDL=False, nest=self._nest
            )
            del fdata
            del temp

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from ..utils import set_numba_threading

from ..timing import function_timer


pysm = None
try:
    import pysm
    import pysm.units as u

    # Disable PySM warnings generated when accessing local
    # data, since these fill the logs.
    import warnings

    warnings.filterwarnings("ignore", category=UserWarning, module="pysm")
except ImportError:
    pysm = None
    u = None


class PySMSky(object):
    """Create a bandpass-integrated sky map with PySM

    Requires PySM 3. It initializes the `pysm.Sky` object either
    in the constructor (`init_sky=True`) or when the `exec` method
    is executed (`init_sky=False`).
    Inizialization of the sky will load templates from the first process
    of the MPI communicator, copy to all processes and then select the local
    rings, distributed as required by `libsharp` for eventual smoothing later.
    If another pixel distribution is required, it can be specified with `pixel_indices`,
    however a different pixel distribution can only be used if performing no smoothing.

    Args:
        comm (mpi4py.MPI.Comm): MPI communicator
        pixels (str): the name of the cache object (<pixels>_<detector>)
            containing the pixel indices to use.
        out (str): accumulate data to the cache with name <out>_<detector>.
            If the named cache objects do not exist, then they are created.
        nside (int): :math:`N_{side}` for PySM
        pysm_sky_config (list(str)): list of PySM components, e.g. ["s1", "d1"],
            this will be passed as `preset_strings` to `pysm.Sky`
        pysm_precomputed_cmb_K_CMB (str): obsolete, will be removed shortly
        pysm_component_objects (list(pysm.Model)): extra sky components that
            inherits from `pysm.Model` to be passed to `pysm.Sky` as `components_objects`
        init_sky (bool): Initializes the sky in the constructor if True, in `exec` if False
        pixel_indices (np.array(int)): List of pixel indices, use None to use the standard
            ring-based libsharp distribution
        units (str): Output units.
    """

    def __init__(
        self,
        comm=None,
        pixels="pixels",
        out="pysm",
        nside=None,
        pysm_sky_config=None,
        pysm_precomputed_cmb_K_CMB=None,
        pysm_component_objects=None,
        init_sky=True,
        pixel_indices=None,
        units="K_CMB",
        map_dist=None,
    ):
        self._nside = nside
        self._pixels = pixels
        self._out = out
        self._pixel_indices = pixel_indices
        self._comm = comm
        self._units = u.Unit(units)
        self.map_dist = map_dist

        self.pysm_sky_config = pysm_sky_config
        self.pysm_precomputed_cmb_K_CMB = pysm_precomputed_cmb_K_CMB
        self.pysm_component_objects = pysm_component_objects
        self.sky = (
            self.init_sky(self.pysm_sky_config, self.pysm_precomputed_cmb_K_CMB)
            if init_sky
            else None
        )

    @function_timer
    def init_sky(self, pysm_sky_config, pysm_precomputed_cmb_K_CMB):
        if pysm is None:
            raise RuntimeError("pysm not available")
        if pysm_precomputed_cmb_K_CMB is not None:
            raise NotImplementedError(
                "pysm_precomputed_cmb_K_CMB is not currently supported"
            )
            # cmb = {
            #     "model": "pre_computed",
            #     "nside": self._nside,
            #     "pixel_indices": self._pixel_indices,
            # }
            # # PySM expects uK_CMB
            # cmb["A_I"], cmb["A_Q"], cmb["A_U"] = (
            #     np.array(
            #         pysm.read_map(
            #             pysm_precomputed_cmb_K_CMB,
            #             self._nside,
            #             field=(0, 1, 2),
            #             pixel_indices=self._pixel_indices,
            #             mpi_comm=self._comm,
            #         )
            #     )
            #     * 1e6
            # )
            # initialized_sky_config["cmb"] = [cmb]
            # # remove cmb from the pysm string
            # pysm_sky_config.pop("cmb", None)
        if self.map_dist is None:
            self.map_dist = (
                None
                if self._comm is None
                else pysm.MapDistribution(
                    pixel_indices=self._pixel_indices,
                    nside=self._nside,
                    mpi_comm=self._comm,
                )
            )
        return pysm.Sky(
            nside=self._nside,
            preset_strings=pysm_sky_config,
            component_objects=self.pysm_component_objects,
            map_dist=self.map_dist,
            output_unit=self._units,
        )

    @function_timer
    def exec(self, local_map, out, bandpasses=None):
        """Execute PySM

        Executes PySM on the given bandpasses and return the
        bandpass-integrated output maps

        Args:
        local_map (dict): Dictionary that will contain output maps
        out (str): Output maps in `local_map` will be named `out_{chname}`
            where `chname` is the key in the bandpasses dictionary
        bandpasses (dict): Dictionary with channel names as keys and
            a tuple of (frequency, weight) as values, PySM will normalize
            the bandpasses and integrate the signal in Jy/sr

        """

        if pysm is None:
            raise RuntimeError("pysm not available")

        if self.sky is None:
            self.sky = self.init_sky(
                self.pysm_sky_config, self.pysm_precomputed_cmb_K_CMB
            )

        for ch_name, bandpass in bandpasses.items():
            out_name = (out + "_" + ch_name) if ch_name else out
            local_map[out_name] = self.sky.get_emission(
                bandpass[0] * u.GHz, bandpass[1]
            ).value
            assert local_map[out_name].shape[0] == 3
            assert local_map[out_name].ndim == 2



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from ..mpi import MPI

import os
import re

import numpy as np

from .. import qarray as qa

from ..utils import Logger
from ..timing import function_timer, Timer

from ..dist import Data, distribute_discrete
from ..op import Operator

from .tod import TOD
from .interval import Interval, intervals_to_chunklist

available = True
try:
    import tidas as tds
    from tidas.mpi import MPIVolume
    from . import tidas_utils as tdsutils
except ImportError:
    available = False


# Module-level constants

STR_QUAT = "QUAT"
STR_FLAG = "FLAG"
STR_COMMON = "COMMON"
STR_BORE = "BORE"
STR_BOREAZEL = "BOREAZEL"
STR_POS = "POS"
STR_VEL = "VEL"
STR_DISTINTR = "chunks"
STR_DETGROUP = "detectors"
STR_FPGROUP = "focalplane"
STR_CACHEGROUP = "cache"
STR_OBSGROUP = "observation"


class TODTidas(TOD):
    """This class provides an interface to a single TIDAS data block.

    An instance of this class reads and writes to a single TIDAS block which
    represents a TOAST observation.  The volume and specific block should
    already exist.  Groups and intervals within the observation may already
    exist or may be created with the provided helper methods.  All groups
    and intervals must exist prior to reading or writing from them.

    Detector pointing offsets from the boresight are given as quaternions,
    and are expected to be contained in the dictionary of properties
    found in the TIDAS group containing detector timestreams.

    This class is generic- it can only read / write standard data streams
    and objects in the TOD.cache.  More specialized uses for specific
    experiments should be implemented in their own classes.

    Args:
        mpicomm (mpi4py.MPI.Comm): the MPI communicator over which this
            observation data is distributed.
        detranks (int):  The dimension of the process grid in the detector
            direction.  The MPI communicator size must be evenly divisible
            by this number.
        vol (tidas.MPIVolume):  The volume.
        path (str):  The path to this observation block.
        group_dets (str):  The name of the TIDAS group containing detector
            and other telescope information at the detector sample rate.
        group_fp (str):  The name of the TIDAS group containing detector
            focalplane offsets.
        detbreaks (list):  Optional list of hard breaks in the detector
            distribution.
        distintervals (str OR list):  Optional name of the TIDAS intervals that
            determines how the data should be distributed along the time
            axis.  If a list is given use that explicit set of sample chunks.

    """

    # FIXME: currently the data flags are stored in the same group as
    # the data.  Once TIDAS supports per-group backend options like
    # compression, we should move the flags to a separate group:
    #   https://github.com/hpc4cmb/tidas/issues/13

    @function_timer
    def __init__(
        self,
        mpicomm,
        detranks,
        vol,
        path,
        group_dets=STR_DETGROUP,
        group_fp=STR_FPGROUP,
        detbreaks=None,
        distintervals=None,
    ):
        if not available:
            raise RuntimeError("tidas is not available")
        rank = 0
        if mpicomm is not None:
            rank = mpicomm.rank

        tmr = Timer()
        tmr.start()

        # We keep a handle to the volume
        self._vol = vol

        # Get a handle to the observation node
        pfields = path.split("/")
        parent = "/".join(pfields[0:-1])
        self._blockparent = parent
        self._blockname = pfields[-1]
        self._blockpath = "/".join([parent, pfields[-1]])
        self._block = None

        # Temporary handle to the block
        tmpblock = tdsutils.find_obs(self._vol, self._blockparent, self._blockname)

        # Get the existing group and interval names
        grpnames = tmpblock.group_names()

        if mpicomm is not None:
            mpicomm.barrier()
        # if rank == 0:
        #     tmr.report_clear("TODTidas open block {}".format(self._blockname))

        # Read detectors and focalplane offsets

        self._fpgrpname = group_fp
        if self._fpgrpname not in grpnames:
            raise RuntimeError(
                "focalplane group {} does not exist".format(self._fpgrpname)
            )

        fpgrp = tmpblock.group_get(self._fpgrpname)
        if fpgrp.size() != 4:
            raise RuntimeError(
                "focalplane group {} does not have quaternions".format(self._fpgrpname)
            )

        self._detquats = dict()
        schm = fpgrp.schema()
        dets = [x.name for x in schm.fields() if (x.name != "TIDAS_TIME")]
        for d in dets:
            q = fpgrp.read(d, 0, 4)
            self._detquats[d] = np.array(q)
        del schm
        del fpgrp

        self._detlist = sorted(list(self._detquats.keys()))
        if mpicomm is not None:
            mpicomm.barrier()
        # if rank == 0:
        #     tmr.report_clear("TODTidas read fp group {}".format(self._blockname))

        # Read detector data group properties and size.

        self._dgrpname = group_dets
        if self._dgrpname not in grpnames:
            raise RuntimeError(
                "detector group {} does not exist".format(self._dgrpname)
            )
        self._dgrp = None

        # Temporary handle to the group
        tmpdgrp = tmpblock.group_get(self._dgrpname)

        self._samples = tmpdgrp.size()
        meta = tdsutils.to_dict(tmpdgrp.dictionary())
        if mpicomm is not None:
            mpicomm.barrier()
        # if rank == 0:
        #     tmr.report_clear(
        #         "TODTidas read det data group {} meta".format(self._blockname)
        #     )

        # See whether we have ground based data

        self._have_azel = False
        schm = tmpdgrp.schema()
        fields = [x.name for x in schm.fields()]
        if "{}_{}X".format(STR_BOREAZEL, STR_QUAT) in fields:
            self._have_azel = True
        del fields
        del schm
        if mpicomm is not None:
            mpicomm.barrier()
        # if rank == 0:
        #     tmr.report_clear("TODTidas check {} azel".format(self._blockname))

        # We need to assign a unique integer index to each detector.  This
        # is used when seeding the streamed RNG in order to simulate
        # timestreams.  For simplicity, and assuming that detector names
        # are not too long, we can convert the detector name to bytes and
        # then to an integer.

        self._detindx = {}
        for det in self._detlist:
            bdet = det.encode("utf-8")
            uid = None
            try:
                ind = int.from_bytes(bdet, byteorder="little")
                uid = int(ind & 0xFFFFFFFF)
            except:
                raise RuntimeError(
                    "Cannot convert detector name {} to a "
                    "unique integer- maybe it is too long?".format(det)
                )
            self._detindx[det] = uid

        if mpicomm is not None:
            mpicomm.barrier()
        # if rank == 0:
        #     tmr.report_clear("TODTidas make {} detindx".format(self._blockname))

        # Create an MPI lock to use for writing to the TIDAS volume.  We must
        # have only one writing process at a time.  Note that this lock is over
        # the communicator for this single observation (TIDAS block).
        # Processes writing to separate blocks have no restrictions.

        # self._writelock = MPILock(mpicomm)

        # Read intervals and set up distribution chunks.

        sampsizes = None

        if distintervals is not None:
            if isinstance(distintervals, str):
                # This is the name of a TIDAS intervals object.  Check that
                # it exists.
                inames = tmpblock.intervals_names()
                if distintervals not in inames:
                    raise RuntimeError(
                        "distintervals {} does not exist".format(distintervals)
                    )
                distint = tmpblock.intervals_get(distintervals)
                # Rank zero process reads and broadcasts intervals
                intervals = None
                if rank == 0:
                    intervals = distint.read()
                if mpicomm is not None:
                    intervals = mpicomm.bcast(intervals, root=0)
                # Compute the contiguous spans of time for data distribution
                # based on the starting points of all intervals.
                sampsizes = intervals_to_chunklist(intervals, self._samples)
                del intervals
                del distint
            else:
                # This must be an explicit list
                sampsizes = list(distintervals)

        if mpicomm is not None:
            mpicomm.barrier()
        # if rank == 0:
        #     tmr.report_clear("TODTidas read {} chunks".format(self._blockname))

        # Delete our temp handles
        del tmpdgrp
        del tmpblock

        # call base class constructor to distribute data
        super().__init__(
            mpicomm,
            self._detlist,
            self._samples,
            detindx=self._detindx,
            detranks=detranks,
            detbreaks=detbreaks,
            sampsizes=sampsizes,
            meta=meta,
        )

        return

    def __del__(self):
        self._close()
        try:
            del self._vol
        except:
            pass
        return

    def _open(self):
        """Open block and group handles."""
        if self._block is not None:
            raise RuntimeError("block is already open!")
        if self._dgrp is not None:
            raise RuntimeError("data group is already open!")
        self._block = tdsutils.find_obs(self._vol, self._blockparent, self._blockname)
        self._dgrp = self._block.group_get(self._dgrpname)
        return

    def _close(self):
        """Close block and group handles."""
        try:
            del self._dgrp
            self._dgrp = None
        except:
            pass
        try:
            del self._block
            self._block = None
        except:
            pass
        return

    @classmethod
    def create(
        cls,
        vol,
        path,
        detectors,
        samples,
        meta,
        azel,
        group_dets=STR_DETGROUP,
        group_fp=STR_FPGROUP,
        units="none",
    ):
        """Class method to create the underlying TIDAS objects.

        This method should ONLY be called by a single process.  It creates the
        detector data group and the detector focalplane offset group.

        Args:
            vol (tidas.MPIVolume):  The volume.
            path (str):  The path to this observation block.
            detectors (dictionary):  Specify the detector offsets.  Each key is
                the detector name, and each value is a quaternion tuple.
            samples (int):  The number of samples to use.
            meta (dict):  Extra scalar key / value parameters to write.
            azel (bool):  If True, this TOD will have ground-based data.
            group_dets (str):  The name of the TIDAS group containing detector
                and other telescope information at the detector sample rate.
            group_fp (str):  The name of the TIDAS group containing detector
                focalplane offsets.
            units (str):  The units of the detector timestreams.

        """
        tmr = Timer()
        tmr.start()

        # Get a handle to the observation node
        pfields = path.split("/")
        parent = "/".join(pfields[0:-1])
        block = tdsutils.find_obs(vol, parent, pfields[-1])
        blockname = pfields[-1]

        # Get the existing group names
        grpnames = block.group_names()
        if group_dets in grpnames:
            raise RuntimeError(
                "Creating new TODTidas, but detector group {} already exists".format(
                    group_dets
                )
            )

        if (detectors is None) or (samples is None):
            raise RuntimeError("detectors and samples must be specified")

        # Focalplane group.  If this group already exists, verify that its
        # data agrees with what we are passing in.

        if group_fp in grpnames:
            fpgrp = block.group_get(group_fp)
            # tmr.report_clear("create {} fp group get".format(blockname))
            for d in sorted(detectors.keys()):
                check = fpgrp.read(d, 0, 4)
                if not np.allclose(detectors[d], check):
                    raise RuntimeError(
                        "existing focalplane offset for {} " "does not match".format(d)
                    )
            # tmr.report_clear("create {} fp group read".format(blockname))
            del fpgrp
        else:
            # Create the FP schema
            schm = cls._create_fp_schema(list(sorted(detectors.keys())))
            # tmr.report_clear("create {} fp schema".format(blockname))

            # Create the FP group
            g = block.group_add(group_fp, tds.Group(schm, tds.Dictionary(), 4))
            # tmr.report_clear("create {} fp group add".format(blockname))

            # Write the FP offsets
            for d in sorted(detectors.keys()):
                g.write(d, 0, np.ascontiguousarray(detectors[d]))

            # tmr.report_clear("create {} fp group write offsets".format(blockname))

            del schm
            del g

        # Detector data group

        # Detector data group properties
        gprops = tdsutils.from_dict(meta)
        # tmr.report_clear("create {} det gprops".format(blockname))

        # Create the detector data schema
        schm = cls._create_det_schema(
            list(sorted(detectors.keys())), tds.DataType.float64, units, azel
        )
        # tmr.report_clear("create {} det schema".format(blockname))

        # Create the detector data group
        g = block.group_add(group_dets, tds.Group(schm, gprops, samples))
        # tmr.report_clear("create {} det group add".format(blockname))

        del schm
        del gprops
        del g
        del block

        return

    @staticmethod
    def _create_det_schema(detlist, datatype, units, azel=False):
        """Create a schema for a list of detectors.

        All detector timestreams will be set to the same type and units.  A
        flag field will be created for each detector.  Additional built-in
        fields will also be added to the schema.

        Args:
            detlist (list): a list of detector names
            datatype (tidas.DataType): the tidas datatype assigned to all
                detector fields.
            units (str): the units string assigned to all data fields.
            azel (bool): if True, data has Az/El pointing as well.

        Returns (tidas.Schema):
            Schema containing the data and flag fields.

        """
        fields = list()
        for c in ["X", "Y", "Z", "W"]:
            f = "{}_{}{}".format(STR_BORE, STR_QUAT, c)
            fields.append(tds.Field(f, tds.DataType.float64, "NA"))
            if azel:
                f = "{}_{}{}".format(STR_BOREAZEL, STR_QUAT, c)
                fields.append(tds.Field(f, tds.DataType.float64, "NA"))

        for c in ["X", "Y", "Z"]:
            f = "{}{}".format(STR_POS, c)
            fields.append(tds.Field(f, tds.DataType.float64, "NA"))

        for c in ["X", "Y", "Z"]:
            f = "{}{}".format(STR_VEL, c)
            fields.append(tds.Field(f, tds.DataType.float64, "NA"))

        fields.append(
            tds.Field("{}_{}".format(STR_FLAG, STR_COMMON), tds.DataType.uint8, "NA")
        )

        for d in detlist:
            fields.append(tds.Field(d, datatype, units))
            fields.append(
                tds.Field("{}_{}".format(STR_FLAG, d), tds.DataType.uint8, "NA")
            )

        return tds.Schema(fields)

    @staticmethod
    def _create_fp_schema(detlist):
        """Create a schema for the focalplane offsets.

        This schema will include one field per detector to hold the
        double-precision quaternion offsets of each detector from the
        boresight.

        Args:
            detlist (list): a list of detector names.

        Returns (tidas.Schema):
            Schema containing the data and flag fields.

        """
        fields = list()
        for d in detlist:
            fields.append(tds.Field(d, tds.DataType.float64, "NA"))
        return tds.Schema(fields)

    @property
    def volume(self):
        """The open TIDAS volume in use by this TOD."""
        return self._vol

    @property
    def block_path(self):
        """The metadata path to this TOD's observation."""
        return self._blockpath

    @property
    def group_det_name(self):
        """The TIDAS group name for the detector data in this TOD."""
        return self._dgrpname

    @property
    def group_fp_name(self):
        """The TIDAS group name for the focalplane data in this TOD."""
        return self._fpgrpname

    def detoffset(self):
        """
        Return dictionary of detector quaternions.

        This returns a dictionary with the detector names as the keys and the
        values are 4-element numpy arrays containing the quaternion offset
        from the boresight.

        Args:
            None

        Returns (dict):
            the dictionary of quaternions.
        """
        return dict(self._detquats)

    @function_timer
    def _read_cache_helper(self, prefix, comps, start, n, usecache):
        """
        Helper function to read multi-component data, pack into an
        array, optionally cache it, and return.
        """
        # Number of components we have
        ncomp = len(comps)

        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start

        if self.cache.exists(prefix):
            return self.cache.reference(prefix)[start : start + n, :]
        else:
            if usecache:
                # We cache the whole observation, regardless of what sample
                # range we will return.
                data = self.cache.create(
                    prefix, np.float64, (self.local_samples[1], ncomp)
                )
                for c in range(ncomp):
                    field = "{}{}".format(prefix, comps[c])
                    d = self._dgrp.read(field, offset, self.local_samples[1])
                    data[:, c] = d
                # Return just the desired slice
                return data[start : start + n, :]
            else:
                # Read and return just the slice we want
                data = np.zeros((n, ncomp), dtype=np.float64)
                for c in range(ncomp):
                    field = "{}{}".format(prefix, comps[c])
                    d = self._dgrp.read(field, offset, n)
                    data[:, c] = d
                return data

    def _write_helper(self, data, prefix, comps, start):
        """
        Helper function to write multi-component data.
        """
        # Number of components we have
        ncomp = len(comps)

        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start

        tmpdata = np.empty(data.shape[0], dtype=data.dtype, order="C")
        for c in range(ncomp):
            field = "{}{}".format(prefix, comps[c])
            tmpdata[:] = data[:, c]
            self._dgrp.write(field, offset, tmpdata)

        return

    def _get_boresight(self, start, n, usecache=True):
        # Cache name
        cachebore = "{}_{}".format(STR_BORE, STR_QUAT)
        # Read and optionally cache the boresight pointing.
        self._open()
        ret = self._read_cache_helper(
            cachebore, ["X", "Y", "Z", "W"], start, n, usecache
        )
        self._close()
        return ret

    def _put_boresight(self, start, data):
        # Data name
        borename = "{}_{}".format(STR_BORE, STR_QUAT)
        # Write data
        # self._writelock.lock()
        self._open()
        self._write_helper(data, borename, ["X", "Y", "Z", "W"], start)
        self._close()
        # self._writelock.unlock()
        return

    def _get_boresight_azel(self, start, n, usecache=True):
        if not self._have_azel:
            raise RuntimeError("No Az/El pointing for this TOD")
        # Cache name
        cachebore = "{}_{}".format(STR_BOREAZEL, STR_QUAT)
        # Read and optionally cache the boresight pointing.
        self._open()
        ret = self._read_cache_helper(
            cachebore, ["X", "Y", "Z", "W"], start, n, usecache
        )
        self._close()
        return ret

    def _put_boresight_azel(self, start, data):
        if not self._have_azel:
            raise RuntimeError("No Az/El pointing for this TOD")
        # Data name
        borename = "{}_{}".format(STR_BOREAZEL, STR_QUAT)
        # Write data
        # self._writelock.lock()
        self._open()
        self._write_helper(data, borename, ["X", "Y", "Z", "W"], start)
        self._close()
        # self._writelock.unlock()
        return

    def _get(self, detector, start, n):
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        # Read from the data group and return
        self._open()
        ret = self._dgrp.read(detector, offset, n)
        self._close()
        return ret

    def _put(self, detector, start, data):
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        # Write to the data group
        # self._writelock.lock()
        self._open()
        self._dgrp.write(detector, offset, np.ascontiguousarray(data))
        self._close()
        # self._writelock.unlock()
        return

    def _get_flags(self, detector, start, n):
        # Field name
        field = "{}_{}".format(STR_FLAG, detector)
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        self._open()
        ret = self._dgrp.read(field, offset, n)
        self._close()
        return ret

    def _put_flags(self, detector, start, flags):
        # Field name
        field = "{}_{}".format(STR_FLAG, detector)
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        # Write to the data group
        # self._writelock.lock()
        self._open()
        self._dgrp.write(field, offset, np.ascontiguousarray(flags))
        self._close()
        # self._writelock.unlock()
        return

    def _get_common_flags(self, start, n):
        # Field name
        field = "{}_{}".format(STR_FLAG, STR_COMMON)
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        # Read from the data group and return
        self._open()
        ret = self._dgrp.read(field, offset, n)
        self._close()
        return ret

    def _put_common_flags(self, start, flags):
        # Field name
        field = "{}_{}".format(STR_FLAG, STR_COMMON)
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        # Write to the data group
        # self._writelock.lock()
        self._open()
        self._dgrp.write(field, offset, np.ascontiguousarray(flags))
        self._close()
        # self._writelock.unlock()
        return

    def _get_times(self, start, n):
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        self._open()
        ret = self._dgrp.read_times(offset, n)
        self._close()
        return ret

    def _put_times(self, start, stamps):
        # Compute the sample offset of our local data
        offset = self.local_samples[0] + start
        # Write to the data group
        # self._writelock.lock()
        self._open()
        self._dgrp.write_times(offset, np.ascontiguousarray(stamps))
        self._close()
        # self._writelock.unlock()
        return

    def _get_pntg(self, detector, start, n):
        # Get boresight pointing (from disk or cache)
        bore = self._get_boresight(start, n)
        # Apply detector quaternion and return
        self._open()
        ret = qa.mult(bore, self._detquats[detector])
        self._close()
        return ret

    def _put_pntg(self, detector, start, data):
        raise RuntimeError(
            "TODTidas computes detector pointing on the fly."
            " Use the write_boresight() method instead."
        )
        return

    def _get_position(self, start, n, usecache=False):
        # Read and optionally cache the telescope position.
        self._open()
        ret = self._read_cache_helper(STR_POS, ["X", "Y", "Z"], start, n, usecache)
        self._close()
        return ret

    def _put_position(self, start, pos):
        # self._writelock.lock()
        self._open()
        self._write_helper(pos, STR_POS, ["X", "Y", "Z"], start)
        self._close()
        # self._writelock.unlock()
        return

    def _get_velocity(self, start, n, usecache=False):
        # Read and optionally cache the telescope velocity.
        self._open()
        ret = self._read_cache_helper(STR_VEL, ["X", "Y", "Z"], start, n, usecache)
        self._close()
        return ret

    def _put_velocity(self, start, vel):
        # self._writelock.lock()
        self._open()
        self._write_helper(vel, STR_VEL, ["X", "Y", "Z"], start)
        self._close()
        # self._writelock.unlock()
        return


@function_timer
def load_tidas(comm, detranks, path, mode, groupname, todclass, **kwargs):
    """Loads an existing TOAST dataset in TIDAS format.

    This takes a 2-level TOAST communicator and opens an existing TIDAS
    volume using the global communicator.  The opened volume handle is stored
    in the observation dictionary with the "tidas" key.  Similarly, the
    metadata path to the block within the volume is stored in the
    "tidas_block" key.

    The leaf nodes of the hierarchy are assumed to be the "observations".
    the observations are assigned to the process groups in a load-balanced
    way based on the number of samples in each detector group.

    For each observation, the specified TOD class is instantiated, and may
    initialize itself using the TIDAS block however it likes.

    Args:
        comm (toast.Comm): the toast Comm class for distributing the data.
        detranks (int):  The dimension of the process grid in the detector
            direction.  The MPI communicator size must be evenly divisible
            by this number.
        path (str):  the TIDAS volume path.
        mode (str): whether to open the file in read-only ("r") or
            read-write ("w") mode.  Default is read-only.
        groupname (str): the name of the group to use for load balancing
            the data distribution.
        todclass (TOD): a TIDAS-compatible TOD class, which must have a
            constructor that takes the MPI communicator, the TIDAS volume,
            the block path, and the size of the process grid in the detector
            direction.  All additional arguments should be keyword args.
        kwargs: All additional arguments to this function are passed to the
            TOD constructor for each observation.

    Returns (toast.Data):
        The distributed data object.

    """
    if not available:
        raise RuntimeError("tidas is not available")
        return None
    # the global communicator
    cworld = comm.comm_world
    # the communicator within the group
    cgroup = comm.comm_group
    # the communicator with all processes with
    # the same rank within their group
    crank = comm.comm_rank

    rank = 0
    if cworld is not None:
        rank = cworld.rank

    # Collectively open the volume.  We cannot use a context manager here,
    # since we are keeping a handle to the volume around for future use.
    # This means the volume will remain open for the life of the program,
    # or (hopefully) will get closed if the distributed data object is
    # destroyed.

    tm = None
    if mode == "w":
        tm = tds.AccessMode.write
    else:
        tm = tds.AccessMode.read
    vol = None
    if cworld is None:
        vol = tds.Volume(path, tm)
    else:
        vol = MPIVolume(cworld, path, tm)

    # Traverse the blocks of the volume and get the properties of the
    # observations so we can distribute them.

    obslist = []
    obsweight = dict()
    obspath = {}

    def procblock(pth, nm, current):
        subs = current.block_names()
        pthnm = "{}/{}".format(pth, nm)
        chld = None
        for s in subs:
            chld = current.block_get(s)
            procblock(pthnm, s, chld)
        if len(subs) == 0:
            # this is a leaf node
            obslist.append(nm)
            obspath[nm] = pthnm
            grpnames = current.group_names()
            if groupname not in grpnames:
                raise RuntimeError(
                    "observation {} does not have a group '{}'".format(nm, groupname)
                )
            grp = current.group_get(groupname)
            sch = grp.schema()
            nfield = len(sch.fields())
            del sch
            obsweight[nm] = nfield * grp.size()
            del grp
        del chld
        return

    if rank == 0:
        root = vol.root()
        toplist = root.block_names()
        bk = None
        for b in toplist:
            bk = root.block_get(b)
            procblock("", b, bk)
        del bk
        del root

    if cworld is not None:
        obslist = cworld.bcast(obslist, root=0)
        obspath = cworld.bcast(obspath, root=0)
        obsweight = cworld.bcast(obsweight, root=0)

    # Distribute observations based on number of samples

    dweight = [obsweight[x] for x in obslist]
    distobs = distribute_discrete(dweight, comm.ngroups)

    # Distributed data

    data = Data(comm)

    # Now every group adds its observations to the list

    firstobs = distobs[comm.group][0]
    nobs = distobs[comm.group][1]
    for ob in range(firstobs, firstobs + nobs):
        obs = dict()
        obs["tod"] = todclass(cgroup, detranks, vol, obspath[obslist[ob]], **kwargs)

        # Get the observation properties group
        pfields = obspath[obslist[ob]].split("/")
        parent = "/".join(pfields[0:-1])
        block = tdsutils.find_obs(vol, parent, pfields[-1])
        obsgroup = block.group_get(STR_OBSGROUP)

        # The observation properties
        obsprops = obsgroup.dictionary()

        props = tdsutils.to_dict(obsprops)
        obspat = re.compile("^obs_(.*)")
        for k, v in props.items():
            obsmat = obspat.match(k)
            if obsmat is not None:
                obs[obsmat.group(1)] = v

        del obsprops
        del obsgroup
        del block

        # Verify that the name on disk is what we expect
        if "name" in obs:
            if obs["name"] != obslist[ob]:
                raise RuntimeError(
                    "loading obs {}: name in file set to {}".format(
                        obslist[ob], obs["name"]
                    )
                )
        else:
            obs["name"] = obslist[ob]

        data.obs.append(obs)

    return data


class OpTidasExport(Operator):
    """Operator which writes data to a TIDAS volume.

    The volume is created at construction time, and the full metadata
    path inside the volume can be given for each observation.  If not given,
    all observations are exported to TIDAS blocks under the root block.

    Timestream data, flags, and boresight pointing are read from the
    current TOD for the observation and written to the TIDAS TOD.  Data can
    be read directly or copied from the cache.

    Args:
        path (str): The output TIDAS volume path.  If this already exists,
            ensure that you specify a group name in the TOD constructor
            options which does not already exist.
        todclass (TOD): a TIDAS-compatible TOD class, which must have a
            constructor that takes the MPI communicator, the TIDAS volume,
            the block path, and the size of the process grid in the detector
            direction.  All additional arguments should be keyword args.  This
            class must also have a class method called "create()", which takes
            arguments for the TIDAS volume, the block path, a dictionary of
            detectors, the number of samples, and a dictionary of metadata.
            All additional arguments should be keyword args.
        backend (str): The TIDAS backend type.  If not specified, the volume
            is assumed to already exist and will be opened for appending.  If
            specified, a new volume will be created.
        comp (str): The TIDAS compression type.  Used only when creating a
            new volume.
        backopts (dict): Extra options to the TIDAS backend.  Used only when
            creating a new volume.
        obspath (dict): (optional) each observation has a "name" and these
            should be the keys of this dictionary.  The values of the dict
            should be the metadata parent path of the observation inside
            the volume.  Default is all observations under the root.
        use_todchunks (bool): if True, use the chunks of the original TOD for
            data distribution.
        use_intervals (bool): if True, use the intervals in the observation
            dictionary for data distribution.
        create_opts (dict): dictionary of options to pass as kwargs to the
            TOD.create() classmethod.
        ctor_opts (dict): dictionary of options to pass as kwargs to the
            TOD constructor.
        cache_name (str):  The name of the cache object (<name>_<detector>) in
            the existing TOD to use for the detector timestream.  If None, use
            the read* methods from the existing TOD.
        cache_common (str):  The name of the cache object in the existing TOD
            to use for common flags.  If None, use the read* methods from the
            existing TOD.
        cache_flag_name (str):   The name of the cache object
            (<name>_<detector>) in the existing TOD to use for the flag
            timestream.  If None, use the read* methods from the existing TOD.

    """

    def __init__(
        self,
        path,
        todclass,
        backend=None,
        comp="none",
        backopts=dict(),
        obspath=None,
        use_todchunks=False,
        use_intervals=False,
        create_opts=dict(),
        ctor_opts=dict(),
        cache_common=None,
        cache_name=None,
        cache_flag_name=None,
    ):

        if not available:
            raise RuntimeError("tidas is not available")

        self._path = path.rstrip("/")
        self._todclass = todclass
        self._backend = None
        if backend is not None:
            if backend == "hdf5":
                self._backend = tds.BackendType.hdf5
            self._comp = None
            if comp == "none":
                self._comp = tds.CompressionType.none
            elif comp == "gzip":
                self._comp = tds.CompressionType.gzip
            elif comp == "bzip2":
                self._comp = tds.CompressionType.bzip2
        self._backopts = backopts
        self._obspath = obspath
        if use_todchunks and use_intervals:
            raise RuntimeError("cannot use both TOD chunks and Intervals")
        self._usechunks = use_todchunks
        self._useintervals = use_intervals
        self._create_opts = create_opts
        self._ctor_opts = ctor_opts
        self._cache_common = cache_common
        self._cache_name = cache_name
        self._cache_flag_name = cache_flag_name
        # We call the parent class constructor
        super().__init__()

    @function_timer
    def exec(self, data):
        """Export data to a TIDAS volume.

        Each group will write its list of observations as TIDAS blocks.

        For errors that prevent the export, this function will directly call
        MPI Abort() rather than raise exceptions.  This could be changed in
        the future if additional logic is implemented to ensure that all
        processes raise an exception when one process encounters an error.

        Args:
            data (toast.Data): The distributed data.

        """
        log = Logger.get()
        # the two-level toast communicator
        comm = data.comm
        # the global communicator
        cworld = comm.comm_world
        # the communicator within the group
        cgroup = comm.comm_group
        # the communicator with all processes with
        # the same rank within their group
        crank = comm.comm_rank

        worldrank = 0
        if cworld is not None:
            worldrank = cworld.rank
        grouprank = 0
        if cgroup is not None:
            grouprank = cgroup.rank

        create = None

        tmr = Timer()
        tmr.start()

        # Handle for the volume.
        vol = None

        # One process checks the path and creates the volume if needed.
        if worldrank == 0:
            if os.path.isdir(self._path):
                # We are appending
                if self._backend is not None:
                    msg = "TIDAS volume {} already exists, but a backend format was specified, which indicates a new volume should be created.".format(
                        self._path
                    )
                    log.error(msg)
                    if cworld is None:
                        raise RuntimeError(msg)
                    else:
                        cworld.Abort()
                create = False
            else:
                # We are creating a new volume.
                if self._backend is None:
                    msg = "TIDAS volume {} does not exist, and a backend format was not specified.".format(
                        self._path
                    )
                    log.error(msg)
                    if cworld is None:
                        raise RuntimeError(msg)
                    else:
                        cworld.Abort()
                create = True
                vol = tds.Volume(self._path, self._backend, self._comp, self._backopts)
                del vol
        if cworld is not None:
            create = cworld.bcast(create, root=0)

        if cworld is not None:
            cworld.barrier()
        if worldrank == 0:
            tmr.report_clear("TIDAS:  Check path / create volume")

        # All processes open the volume.  Note:  we *might* be exporting a new
        # detector group from a toast.Data instance that is already using this
        # volume.  In this case, we need to not modify existing metadata in
        # the volume and only add our new group.

        if cworld is None:
            vol = tds.Volume(self._path, tds.AccessMode.write)
        else:
            vol = MPIVolume(cworld, self._path, tds.AccessMode.write)

        if cworld is not None:
            cworld.barrier()
        if worldrank == 0:
            tmr.report_clear("TIDAS:  world open volume")

        # The rank zero process in each group creates or checks the observation
        # blocks and creates the TOD groups.

        if grouprank == 0:
            for obs in data.obs:
                # The existing TOD
                oldtod = obs["tod"]
                oldcomm = oldtod.mpicomm

                # Sanity check- the group communicator should be the same
                different_comm = False
                if cgroup is None:
                    if oldcomm is not None:
                        different_comm = True
                else:
                    if oldcomm is None:
                        different_comm = True
                    else:
                        comp = MPI.Comm.Compare(oldtod.mpicomm, cgroup)
                        if comp not in (MPI.IDENT, MPI.CONGRUENT):
                            different_comm = True
                if different_comm:
                    msg = "On export, original TOD comm ({}) is different from group comm ({}).  Comm compare = {}".format(
                        oldcomm, cgroup, comp
                    )
                    log.error(msg)
                    if cgroup is None:
                        raise RuntimeError(msg)
                    else:
                        cworld.Abort()

                # Get the name
                if "name" not in obs:
                    log.error("observation does not have a name, cannot export")
                    if cgroup is None:
                        raise RuntimeError()
                    else:
                        cworld.Abort()
                obsname = obs["name"]

                # Get the metadata path
                blockpath = ""
                if self._obspath is not None:
                    blockpath = self._obspath[obsname]

                detranks, sampranks = oldtod.grid_size
                rankdet, ranksamp = oldtod.grid_ranks

                # Either create the observation group, or verify that the
                # contents are consistent.

                # Get any scalar metadata in the existing observation dictionary
                oprops = dict()
                for k, v in obs.items():
                    if isinstance(v, (bool, int, float, str)):
                        oprops["obs_{}".format(k)] = v

                # Create or open the block in the volume that corresponds to
                # this observation.
                tob = tdsutils.find_obs(vol, blockpath, obsname)

                obsprops = None
                obsgroup = None

                if create:
                    # The observation properties
                    obsprops = tdsutils.from_dict(oprops)

                    # Create the observation group
                    obsgroup = tob.group_add(
                        STR_OBSGROUP, tds.Group(tds.Schema(), obsprops, 0)
                    )
                else:
                    # Get the observation properties group
                    obsgroup = tob.group_get(STR_OBSGROUP)

                    # The observation properties
                    obsprops = obsgroup.dictionary()

                    checkprops = tdsutils.to_dict(obsprops)
                    obspat = re.compile("^obs_(.*)")
                    for k, v in checkprops.items():
                        obsmat = obspat.match(k)
                        if obsmat is not None:
                            obskey = obsmat.group(1)
                            if (obskey not in obs) or (obs[obskey] != v):
                                raise RuntimeError(
                                    "Attempting to export to existing observation with missing / changed property '{}'".format(
                                        k
                                    )
                                )

                # Whatever intervals we are using for data distribution (TOD
                # chunks or the input observation intervals), we want to write
                # those to the TIDAS volume.

                ninterval = None
                if self._useintervals:
                    # List of current toast intervals
                    if "intervals" not in obs:
                        raise RuntimeError("Observation does not contain " "intervals")
                    ninterval = len(obs["intervals"])
                else:
                    # We are using TOD chunks
                    ninterval = len(oldtod.total_chunks)

                allintr = tob.intervals_names()
                if STR_DISTINTR in allintr:
                    # Intervals already exist- verify size
                    tintr = tob.intervals_get(STR_DISTINTR)
                    if tintr.size() != ninterval:
                        raise RuntimeError(
                            "Existing intervals have different "
                            "length than ones being exported"
                        )
                    del tintr
                else:
                    # Create distribution intervals
                    tintr = tob.intervals_add(
                        STR_DISTINTR, tds.Intervals(tds.Dictionary(), ninterval)
                    )
                    del tintr

                del obsgroup
                del obsprops
                del tob

                # Do we have ground pointing?
                azel = None
                try:
                    test = oldtod.read_boresight_azel(local_start=0, n=1)
                    azel = True
                    del test
                except:
                    azel = False

                self._todclass.create(
                    vol,
                    "/".join([blockpath, obsname]),
                    oldtod.detoffset(),
                    oldtod.total_samples,
                    oldtod.meta,
                    azel,
                    **self._create_opts
                )

        if cworld is not None:
            cworld.barrier()
        if worldrank == 0:
            tmr.report_clear("TIDAS:  Create obs groups and intervals")

        # print("group {} meta sync".format(comm.group), flush=True)
        if cworld is not None:
            vol.meta_sync()
        del vol

        # Now all metadata is written to disk and synced between processes.
        if cworld is not None:
            cworld.barrier()
        if worldrank == 0:
            tmr.report_clear("TIDAS:  World volume sync")

        # Every process group copies their TOD data for each observation.

        if cworld is None:
            vol = tds.Volume(self._path, tds.AccessMode.write)
        else:
            vol = MPIVolume(cgroup, self._path, tds.AccessMode.write)

        for obs in data.obs:
            # Get the name
            obsname = obs["name"]

            # Get the metadata path
            blockpath = ""
            if self._obspath is not None:
                blockpath = self._obspath[obsname]

            tob = tdsutils.find_obs(vol, blockpath, obsname)

            # The existing TOD
            oldtod = obs["tod"]
            detranks, sampranks = oldtod.grid_size
            rankdet, ranksamp = oldtod.grid_ranks

            # Write out out distribution intervals.  This is either from the
            # old TOD chunks (combined with timestamp information), or using
            # existing intervals from the old observation.

            if self._useintervals:
                # We are using the observation intervals (every process has
                # a copy of these).  The root process writes the data.
                if cgroup.rank == 0:
                    ilist = []
                    for intr in obs["intervals"]:
                        ilist.append(
                            tds.Intrvl(intr.start, intr.stop, intr.first, intr.last)
                        )
                    tintr = tob.intervals_get(STR_DISTINTR)
                    tintr.write(ilist)
                    del tintr
                    del ilist
                if cgroup is not None:
                    cgroup.barrier()
                if grouprank == 0:
                    tmr.report_clear(
                        "TIDAS:  Group write obs {} intervals".format(obsname)
                    )
            else:
                # We are using the TOD chunks and the timestamps to build
                # the intervals.  Gather the timestamps to the root process
                # and do this write on that process.
                if rankdet == 0:
                    psize = oldtod.local_samples[1]
                    pdata = oldtod.read_times()
                    psizes = None
                    if oldtod.grid_comm_row is None:
                        psizes = [psize]
                    else:
                        psizes = oldtod.grid_comm_row.gather(psize, root=0)
                    disp = None
                    stamps = None
                    if ranksamp == 0:
                        # Compute the displacements into the receive buffer.
                        disp = [0]
                        for ps in psizes[:-1]:
                            last = disp[-1]
                            disp.append(last + ps)
                        # allocate receive buffer
                        stamps = np.zeros(np.sum(psizes), dtype=np.float64)

                    if oldtod.grid_comm_row is None:
                        stamps[:] = pdata
                    else:
                        oldtod.grid_comm_row.Gatherv(
                            pdata, [stamps, psizes, disp, MPI.DOUBLE], root=0
                        )

                    if ranksamp == 0:
                        ilist = []
                        ckoff = 0
                        for chk in oldtod.total_chunks:
                            ifirst = ckoff
                            ilast = ckoff + chk - 1
                            istart = stamps[ifirst]
                            istop = stamps[ilast]
                            ilist.append(tds.Intrvl(istart, istop, ifirst, ilast))
                            ckoff += chk
                        tintr = tob.intervals_get(STR_DISTINTR)
                        tintr.write(ilist)
                        del tintr
                        del ilist
                if cgroup is not None:
                    cgroup.barrier()
                if grouprank == 0:
                    tmr.report_clear(
                        "TIDAS:  Group write {} chunk intervals".format(obsname)
                    )

            del tob

            # Determine data distribution chunks
            sampsizes = None
            if self._usechunks:
                sampsizes = oldtod.total_chunks
            elif self._useintervals:
                sampsizes = intervals_to_chunklist(
                    obs["intervals"], oldtod.total_samples
                )

            if cgroup is not None:
                cgroup.barrier()
            if grouprank == 0:
                tmr.report_clear("TIDAS:  Group compute {} sampsizes".format(obsname))

            # The new TIDAS TOD.  Note:  the TOD instance will maintain
            # a handle on the volume for its lifetime.

            obsdir = "/".join([blockpath, obsname])

            # print("export:  construct tod at {}".format(obsdir), flush=True)
            tod = self._todclass(
                cgroup,
                detranks,
                vol,
                obsdir,
                distintervals=sampsizes,
                **self._ctor_opts
            )
            # print("export:  done construct tod at {}".format(obsdir), flush=True)

            if cgroup is not None:
                cgroup.barrier()
            if grouprank == 0:
                tmr.report_clear("TIDAS:  Group instantiate {} TOD".format(obsname))

            # Sanity check that the process grids are the same
            new_detranks, new_sampranks = tod.grid_size
            new_rankdet, new_ranksamp = tod.grid_ranks

            if (
                (new_detranks != detranks)
                or (new_sampranks != sampranks)
                or (new_rankdet != rankdet)
                or (new_ranksamp != ranksamp)
            ):
                msg = "TIDAS: During export to obs ({}), process grid shape mismatch".format(
                    obsname
                )
                log.error(msg)
                if cgroup is None:
                    raise RuntimeError(msg)
                else:
                    cgroup.Abort()

            if cgroup is not None:
                cgroup.barrier()
            if grouprank == 0:
                tmr.report_clear(
                    "TIDAS:  Group check TOD {} process grid".format(obsname)
                )

            # Some data is common across all processes that share the same
            # time span (timestamps, boresight pointing, common flags).
            # Since we only need to write these once, we let the first
            # process row handle that.

            # print("export:  start data copy for {}".format(obsdir), flush=True)

            if rankdet == 0:
                # Only the first row of the process grid does this.
                for p in range(sampranks):
                    # Processes take turns in the sample direction.
                    if p == ranksamp:
                        tod.write_times(stamps=oldtod.read_times())
                        tod.write_boresight(data=oldtod.read_boresight())
                        if azel:
                            tod.write_boresight_azel(data=oldtod.read_boresight_azel())
                        tod.write_position(pos=oldtod.read_position())
                        tod.write_velocity(vel=oldtod.read_velocity())
                        if self._cache_common is None:
                            tod.write_common_flags(flags=oldtod.read_common_flags())
                        else:
                            ref = oldtod.cache.reference(self._cache_common)
                            tod.write_common_flags(flags=ref)
                            del ref
                    if tod.grid_comm_row is not None:
                        tod.grid_comm_row.barrier()

            if cgroup is not None:
                cgroup.barrier()
            if grouprank == 0:
                tmr.report_clear(
                    "TIDAS:  Group write {} timestamps and boresight".format(obsname)
                )

            # print("export:  finish common copy for {}".format(obsdir), flush=True)

            # Now every process takes turns writing their unique data
            groupsize = 1
            if cgroup is not None:
                groupsize = cgroup.size
            for p in range(groupsize):
                if p == grouprank:
                    for d in oldtod.local_dets:
                        if self._cache_name is None:
                            tod.write(detector=d, data=oldtod.read(detector=d))
                        else:
                            ref = oldtod.cache.reference(
                                "{}_{}".format(self._cache_name, d)
                            )
                            tod.write(detector=d, data=ref)
                            del ref
                        if self._cache_flag_name is None:
                            tod.write_flags(
                                detector=d, flags=oldtod.read_flags(detector=d)
                            )
                        else:
                            ref = oldtod.cache.reference(
                                "{}_{}".format(self._cache_flag_name, d)
                            )
                            tod.write_flags(detector=d, flags=ref)
                            del ref
                if cgroup is not None:
                    cgroup.barrier()

            if cgroup is not None:
                cgroup.barrier()
            if grouprank == 0:
                tmr.report_clear("TIDAS:  Group write {} data".format(obsname))

            del tod

        del vol

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

_matplotlib_backend = None


def set_backend(backend="agg"):
    global _matplotlib_backend
    if _matplotlib_backend is None:
        _matplotlib_backend = backend
        import matplotlib

        matplotlib.use(_matplotlib_backend)
    return
